@@ -1,20 +1,10 @@
+// SPDX-License-Identifier: LGPL-2.1-or-later
 /*
  * libiio - Library for interfacing industrial I/O (IIO) devices
  *
  * Copyright (C) 2014 Analog Devices, Inc.
  * Author: Paul Cercueil <paul.cercueil@analog.com>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * */
+ */
 
 #include "debug.h"
 #include "iio-private.h"
@@ -51,6 +41,26 @@ static const char * const iio_chan_type_name_spec[] = {
 	[IIO_CONCENTRATION] = "concentration",
 	[IIO_RESISTANCE] = "resistance",
 	[IIO_PH] = "ph",
+	[IIO_UVINDEX] = "uvindex",
+	[IIO_ELECTRICALCONDUCTIVITY] = "electricalconductivity",
+	[IIO_COUNT] = "count",
+	[IIO_INDEX] = "index",
+	[IIO_GRAVITY] = "gravity",
+	[IIO_POSITIONRELATIVE] = "positionrelative",
+	[IIO_PHASE] = "phase",
+	[IIO_MASSCONCENTRATION] = "massconcentration",
+};
+
+static const char * const hwmon_chan_type_name_spec[] = {
+	[HWMON_VOLTAGE] = "in",
+	[HWMON_FAN] = "fan",
+	[HWMON_PWM] = "pwm",
+	[HWMON_TEMP] = "temp",
+	[HWMON_CURRENT] = "curr",
+	[HWMON_POWER] = "power",
+	[HWMON_ENERGY] = "energy",
+	[HWMON_HUMIDITY] = "humidity",
+	[HWMON_INTRUSION] = "intrusion",
 };
 
 static const char * const modifier_names[] = {
@@ -73,6 +83,8 @@ static const char * const modifier_names[] = {
 	[IIO_MOD_LIGHT_RED] = "red",
 	[IIO_MOD_LIGHT_GREEN] = "green",
 	[IIO_MOD_LIGHT_BLUE] = "blue",
+	[IIO_MOD_LIGHT_UV] = "uv",
+	[IIO_MOD_LIGHT_DUV] = "duv",
 	[IIO_MOD_QUATERNION] = "quaternion",
 	[IIO_MOD_TEMP_AMBIENT] = "ambient",
 	[IIO_MOD_TEMP_OBJECT] = "object",
@@ -88,7 +100,14 @@ static const char * const modifier_names[] = {
 	[IIO_MOD_I] = "i",
 	[IIO_MOD_Q] = "q",
 	[IIO_MOD_CO2] = "co2",
+	[IIO_MOD_ETHANOL] = "ethanol",
+	[IIO_MOD_H2] = "h2",
+	[IIO_MOD_O2] = "o2",
 	[IIO_MOD_VOC] = "voc",
+	[IIO_MOD_PM1] = "pm1",
+	[IIO_MOD_PM2P5] = "pm2p5",
+	[IIO_MOD_PM4] = "pm4",
+	[IIO_MOD_PM10] = "pm10",
 };
 
 /*
@@ -117,6 +136,36 @@ unsigned int find_channel_modifier(const char *s, size_t *len_p)
 	return IIO_NO_MOD;
 }
 
+static int iio_channel_find_type(const char *id,
+			const char *const *name_spec, size_t size)
+{
+	unsigned int i;
+	size_t len;
+
+	for (i = 0; i < size; i++) {
+		len = strlen(name_spec[i]);
+		if (strncmp(name_spec[i], id, len) != 0)
+		      continue;
+
+		/* Type must be followed by one of a '\0', a '_', or a digit */
+		if (id[len] != '\0' && id[len] != '_' &&
+				(id[len] < '0' || id[len] > '9'))
+			continue;
+
+		return i;
+	}
+
+	return -EINVAL;
+}
+
+#if WITH_HWMON
+bool iio_channel_is_hwmon(const char *id)
+{
+	return iio_channel_find_type(id, hwmon_chan_type_name_spec,
+				ARRAY_SIZE(hwmon_chan_type_name_spec)) >= 0;
+}
+#endif /* WITH_HWMON */
+
 /*
  * Initializes all auto-detected fields of the channel struct. Must be called
  * after the channel has been otherwise fully initialized.
@@ -126,21 +175,19 @@ void iio_channel_init_finalize(struct iio_channel *chn)
 	unsigned int i;
 	size_t len;
 	char *mod;
+	int type;
 
-	chn->type = IIO_CHAN_TYPE_UNKNOWN;
-	chn->modifier = IIO_NO_MOD;
-
-	for (i = 0; i < ARRAY_SIZE(iio_chan_type_name_spec); i++) {
-		len = strlen(iio_chan_type_name_spec[i]);
-		if (strncmp(iio_chan_type_name_spec[i], chn->id, len) != 0)
-			continue;
-		/* Type must be followed by either a '_' or a digit */
-		if (chn->id[len] != '_' && chn->id[len] < '0' && chn->id[len] > '9')
-			continue;
-
-		chn->type = (enum iio_chan_type) i;
+	if (iio_device_is_hwmon(chn->dev)) {
+		type = iio_channel_find_type(chn->id, hwmon_chan_type_name_spec,
+					ARRAY_SIZE(hwmon_chan_type_name_spec));
+	} else {
+		type = iio_channel_find_type(chn->id, iio_chan_type_name_spec,
+					ARRAY_SIZE(iio_chan_type_name_spec));
 	}
 
+	chn->type = (type >= 0) ? type : IIO_CHAN_TYPE_UNKNOWN;
+	chn->modifier = IIO_NO_MOD;
+
 	mod = strchr(chn->id, '_');
 	if (!mod)
 		return;
@@ -153,140 +200,98 @@ void iio_channel_init_finalize(struct iio_channel *chn)
 		len = strlen(modifier_names[i]);
 		if (strncmp(modifier_names[i], mod, len) != 0)
 			continue;
-		/* Modifier must be followed by a '_' */
-		if (mod[len] != '_')
-			continue;
 
 		chn->modifier = (enum iio_modifier) i;
 		break;
 	}
 }
 
-static char *get_attr_xml(struct iio_channel_attr *attr, size_t *length)
+static ssize_t iio_snprintf_chan_attr_xml(char *str, ssize_t len,
+					  struct iio_channel_attr *attr)
 {
-	char *str;
-	size_t len = strlen(attr->name) + sizeof("<attribute name=\"\" />");
-	if (attr->filename)
-		len += strlen(attr->filename) + sizeof("filename=\"\"");
+	ssize_t ret, alen = 0;
 
-	str = malloc(len);
-	if (!str)
-		return NULL;
+	if (!attr->filename)
+		return iio_snprintf(str, len, "<attribute name=\"%s\" />", attr->name);
 
-	*length = len - 1; /* Skip the \0 */
-	if (attr->filename)
-		iio_snprintf(str, len, "<attribute name=\"%s\" filename=\"%s\" />",
-				attr->name, attr->filename);
-	else
-		iio_snprintf(str, len, "<attribute name=\"%s\" />", attr->name);
-	return str;
+	ret = iio_snprintf(str, len, "<attribute name=\"%s\" ", attr->name);
+	if (ret < 0)
+		return ret;
+
+	iio_update_xml_indexes(ret, &str, &len, &alen);
+
+	ret = iio_xml_print_and_sanitized_param(str, len, "filename=\"",
+						attr->filename, "\" />");
+	if (ret < 0)
+		return ret;
+
+	return alen + ret;
 }
 
-static char * get_scan_element(const struct iio_channel *chn, size_t *length)
+static ssize_t iio_snprintf_scan_element_xml(char *str, ssize_t len,
+					     const struct iio_channel *chn)
 {
-	char buf[1024], repeat[8] = "", *str;
 	char processed = (chn->format.is_fully_defined ? 'A' - 'a' : 0);
+	char repeat[12] = "", scale[48] = "";
 
 	if (chn->format.repeat > 1)
 		iio_snprintf(repeat, sizeof(repeat), "X%u", chn->format.repeat);
 
-	iio_snprintf(buf, sizeof(buf), "<scan-element index=\"%li\" "
-			"format=\"%ce:%c%u/%u%s&gt;&gt;%u\" />",
+	if (chn->format.with_scale)
+		iio_snprintf(scale, sizeof(scale), "scale=\"%f\" ", chn->format.scale);
+
+	return iio_snprintf(str, len,
+			"<scan-element index=\"%li\" format=\"%ce:%c%u/%u%s&gt;&gt;%u\" %s/>",
 			chn->index, chn->format.is_be ? 'b' : 'l',
 			chn->format.is_signed ? 's' + processed : 'u' + processed,
 			chn->format.bits, chn->format.length, repeat,
-			chn->format.shift);
-
-	if (chn->format.with_scale) {
-		char *ptr = strrchr(buf, '\0');
-		iio_snprintf(ptr - 2, buf + sizeof(buf) - ptr + 2,
-				"scale=\"%f\" />", chn->format.scale);
-	}
-
-	str = iio_strdup(buf);
-	if (str)
-		*length = strlen(str);
-	return str;
+			chn->format.shift, scale);
 }
 
-/* Returns a string containing the XML representation of this channel */
-char * iio_channel_get_xml(const struct iio_channel *chn, size_t *length)
+ssize_t iio_snprintf_channel_xml(char *ptr, ssize_t len,
+				 const struct iio_channel *chn)
 {
-	size_t len = sizeof("<channel id=\"\" name=\"\" "
-			"type=\"output\" ></channel>")
-		+ strlen(chn->id) + (chn->name ? strlen(chn->name) : 0);
-	char *ptr, *str, **attrs, *scan_element = NULL;
-	size_t *attrs_len, scan_element_len = 0;
+	ssize_t ret, alen = 0;
 	unsigned int i;
 
-	if (chn->is_scan_element) {
-		scan_element = get_scan_element(chn, &scan_element_len);
-		if (!scan_element)
-			return NULL;
-		else
-			len += scan_element_len;
-	}
 
-	attrs_len = malloc(chn->nb_attrs * sizeof(*attrs_len));
-	if (!attrs_len)
-		goto err_free_scan_element;
-
-	attrs = malloc(chn->nb_attrs * sizeof(*attrs));
-	if (!attrs)
-		goto err_free_attrs_len;
-
-	for (i = 0; i < chn->nb_attrs; i++) {
-		char *xml = get_attr_xml(&chn->attrs[i], &attrs_len[i]);
-		if (!xml)
-			goto err_free_attrs;
-		attrs[i] = xml;
-		len += attrs_len[i];
-	}
-
-	str = malloc(len);
-	if (!str)
-		goto err_free_attrs;
-
-	iio_snprintf(str, len, "<channel id=\"%s\"", chn->id);
-	ptr = strrchr(str, '\0');
+	ret = iio_xml_print_and_sanitized_param(ptr, len, "<channel id=\"",
+						chn->id, "\"");
+	if (ret < 0)
+		return ret;
+	iio_update_xml_indexes(ret, &ptr, &len, &alen);
 
 	if (chn->name) {
-		sprintf(ptr, " name=\"%s\"", chn->name);
-		ptr = strrchr(ptr, '\0');
+		ret = iio_snprintf(ptr, len, " name=\"%s\"", chn->name);
+		if (ret < 0)
+			return ret;
+		iio_update_xml_indexes(ret, &ptr, &len, &alen);
 	}
 
-	sprintf(ptr, " type=\"%s\" >", chn->is_output ? "output" : "input");
-	ptr = strrchr(ptr, '\0');
+	ret = iio_snprintf(ptr, len, " type=\"%s\" >", chn->is_output ? "output" : "input");
+	if (ret < 0)
+		return ret;
+	iio_update_xml_indexes(ret, &ptr, &len, &alen);
 
 	if (chn->is_scan_element) {
-		strcpy(ptr, scan_element);
-		ptr += scan_element_len;
+		ret = iio_snprintf_scan_element_xml(ptr, len, chn);
+		if (ret < 0)
+			return ret;
+		iio_update_xml_indexes(ret, &ptr, &len, &alen);
 	}
 
 	for (i = 0; i < chn->nb_attrs; i++) {
-		strcpy(ptr, attrs[i]);
-		ptr += attrs_len[i];
-		free(attrs[i]);
+		ret = iio_snprintf_chan_attr_xml(ptr, len, &chn->attrs[i]);
+		if (ret < 0)
+			return ret;
+		iio_update_xml_indexes(ret, &ptr, &len, &alen);
 	}
 
-	free(scan_element);
-	free(attrs);
-	free(attrs_len);
+	ret = iio_snprintf(ptr, len, "</channel>");
+	if (ret < 0)
+		return ret;
 
-	strcpy(ptr, "</channel>");
-	*length = ptr - str + sizeof("</channel>") - 1;
-	return str;
-
-err_free_attrs:
-	while (i--)
-		free(attrs[i]);
-	free(attrs);
-err_free_attrs_len:
-	free(attrs_len);
-err_free_scan_element:
-	if (chn->is_scan_element)
-		free(scan_element);
-	return NULL;
+	return alen + ret;
 }
 
 const char * iio_channel_get_id(const struct iio_channel *chn)
@@ -417,12 +422,9 @@ void free_channel(struct iio_channel *chn)
 		free(chn->attrs[i].name);
 		free(chn->attrs[i].filename);
 	}
-	if (chn->nb_attrs)
-		free(chn->attrs);
-	if (chn->name)
-		free(chn->name);
-	if (chn->id)
-		free(chn->id);
+	free(chn->attrs);
+	free(chn->name);
+	free(chn->id);
 	free(chn);
 }
 
@@ -438,11 +440,7 @@ static void shift_bits(uint8_t *dst, size_t shift, size_t len, bool left)
 	size_t i, shift_bytes = shift / 8;
 	shift %= 8;
 
-#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
-	if (!left)
-#else
-	if (left)
-#endif
+	if (is_little_endian() ^ left)
 	{
 		if (shift_bytes) {
 			memmove(dst, dst + shift_bytes, len - shift_bytes);
@@ -450,15 +448,15 @@ static void shift_bits(uint8_t *dst, size_t shift, size_t len, bool left)
 		}
 		if (shift) {
 			for (i = 0; i < len; i++) {
-#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
-				dst[i] >>= shift;
-				if (i < len - 1)
-					dst[i] |= dst[i + 1] << (8 - shift);
-#else
-				dst[i] <<= shift;
-				if (i < len - 1)
-					dst[i] |= dst[i + 1] >> (8 - shift);
-#endif
+				if (is_little_endian()) {
+					dst[i] >>= shift;
+					if (i < len - 1)
+						dst[i] |= dst[i + 1] << (8 - shift);
+				} else {
+					dst[i] <<= shift;
+					if (i < len - 1)
+						dst[i] |= dst[i + 1] >> (8 - shift);
+				}
 			}
 		}
 	} else {
@@ -468,15 +466,15 @@ static void shift_bits(uint8_t *dst, size_t shift, size_t len, bool left)
 		}
 		if (shift) {
 			for (i = len; i > 0; i--) {
-#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
-				dst[i - 1] <<= shift;
-				if (i > 1)
-					dst[i - 1] |= dst[i - 2] >> (8 - shift);
-#else
-				dst[i - 1] >>= shift;
-				if (i > 1)
-					dst[i - 1] |= dst[i - 2] << (8 - shift);
-#endif
+				if (is_little_endian()) {
+					dst[i - 1] <<= shift;
+					if (i > 1)
+						dst[i - 1] |= dst[i - 2] >> (8 - shift);
+				} else {
+					dst[i - 1] >>= shift;
+					if (i > 1)
+						dst[i - 1] |= dst[i - 2] << (8 - shift);
+				}
 			}
 		}
 	}
@@ -487,22 +485,22 @@ static void sign_extend(uint8_t *dst, size_t bits, size_t len)
 	size_t upper_bytes = ((len * 8 - bits) / 8);
 	uint8_t msb, msb_bit = 1 << ((bits - 1) % 8);
 
-#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
-	msb = dst[len - 1 - upper_bytes] & msb_bit;
-	if (upper_bytes)
-		memset(dst + len - upper_bytes, msb ? 0xff : 0x00, upper_bytes);
-	if (msb)
-		dst[len - 1 - upper_bytes] |= ~(msb_bit - 1);
-	else
-		dst[len - 1 - upper_bytes] &= (msb_bit - 1);
-#else
-	/* XXX: untested */
-	msb = dst[upper_bytes] & msb_bit;
-	if (upper_bytes)
-		memset(dst, msb ? 0xff : 0x00, upper_bytes);
-	if (msb)
-		dst[upper_bytes] |= ~(msb_bit - 1);
-#endif
+	if (is_little_endian()) {
+		msb = dst[len - 1 - upper_bytes] & msb_bit;
+		if (upper_bytes)
+			memset(dst + len - upper_bytes, msb ? 0xff : 0x00, upper_bytes);
+		if (msb)
+			dst[len - 1 - upper_bytes] |= ~(msb_bit - 1);
+		else
+			dst[len - 1 - upper_bytes] &= (msb_bit - 1);
+	} else {
+		/* XXX: untested */
+		msb = dst[upper_bytes] & msb_bit;
+		if (upper_bytes)
+			memset(dst, msb ? 0xff : 0x00, upper_bytes);
+		if (msb)
+			dst[upper_bytes] |= ~(msb_bit - 1);
+	}
 }
 
 static void mask_upper_bits(uint8_t *dst, size_t bits, size_t len)
@@ -526,11 +524,7 @@ void iio_channel_convert(const struct iio_channel *chn,
 	unsigned int len = chn->format.length / 8;
 	ptrdiff_t end = len * chn->format.repeat;
 	uintptr_t end_ptr = src_ptr + end;
-#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
-	bool swap = chn->format.is_be;
-#else
-	bool swap = !chn->format.is_be;
-#endif
+	bool swap = is_little_endian() ^ !chn->format.is_be;
 
 	for (src_ptr = (uintptr_t) src; src_ptr < end_ptr;
 			src_ptr += len, dst_ptr += len) {
@@ -562,11 +556,7 @@ void iio_channel_convert_inverse(const struct iio_channel *chn,
 	unsigned int len = chn->format.length / 8;
 	ptrdiff_t end = len * chn->format.repeat;
 	uintptr_t end_ptr = dst_ptr + end;
-#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
-	bool swap = chn->format.is_be;
-#else
-	bool swap = !chn->format.is_be;
-#endif
+	bool swap = is_little_endian() ^ !chn->format.is_be;
 	uint8_t buf[1024];
 
 	/* Somehow I doubt we will have samples of 8192 bits each. */
@@ -659,8 +649,9 @@ int iio_channel_attr_read_longlong(const struct iio_channel *chn,
 	if (ret < 0)
 		return (int) ret;
 
+	errno = 0;
 	value = strtoll(buf, &end, 0);
-	if (end == buf)
+	if (end == buf || errno == ERANGE)
 		return -EINVAL;
 	*val = value;
 	return 0;
@@ -692,33 +683,33 @@ int iio_channel_attr_read_double(const struct iio_channel *chn,
 int iio_channel_attr_write_longlong(const struct iio_channel *chn,
 		const char *attr, long long val)
 {
-	ssize_t ret;
+	int ret;
 	char buf[1024];
 	iio_snprintf(buf, sizeof(buf), "%lld", val);
-	ret = iio_channel_attr_write(chn, attr, buf);
+	ret = (int) iio_channel_attr_write(chn, attr, buf);
 	return ret < 0 ? ret : 0;
 }
 
 int iio_channel_attr_write_double(const struct iio_channel *chn,
 		const char *attr, double val)
 {
-	ssize_t ret;
+	int ret;
 	char buf[1024];
 
-	ret = (ssize_t) write_double(buf, sizeof(buf), val);
+	ret = write_double(buf, sizeof(buf), val);
 	if (!ret)
-		ret = iio_channel_attr_write(chn, attr, buf);
+		ret = (int) iio_channel_attr_write(chn, attr, buf);
 	return ret < 0 ? ret : 0;
 }
 
 int iio_channel_attr_write_bool(const struct iio_channel *chn,
 		const char *attr, bool val)
 {
-	ssize_t ret;
+	int ret;
 	if (val)
-		ret = iio_channel_attr_write_raw(chn, attr, "1", 2);
+		ret = (int) iio_channel_attr_write_raw(chn, attr, "1", 2);
 	else
-		ret = iio_channel_attr_write_raw(chn, attr, "0", 2);
+		ret = (int) iio_channel_attr_write_raw(chn, attr, "0", 2);
 	return ret < 0 ? ret : 0;
 }
 
@@ -738,7 +729,7 @@ int iio_channel_attr_read_all(struct iio_channel *chn,
 			const char *attr, const char *val, size_t len, void *d),
 		void *data)
 {
-	int ret;
+	int ret, buf_size;
 	char *buf, *ptr;
 	unsigned int i;
 
@@ -752,12 +743,26 @@ int iio_channel_attr_read_all(struct iio_channel *chn,
 		goto err_free_buf;
 
 	ptr = buf;
+	buf_size = ret;
 
 	for (i = 0; i < iio_channel_get_attrs_count(chn); i++) {
 		const char *attr = iio_channel_get_attr(chn, i);
-		int32_t len = (int32_t) iio_be32toh(*(uint32_t *) ptr);
+		int32_t len;
 
+		if (buf_size < 4) {
+			ret = -EPROTO;
+			break;
+		}
+
+		len = (int32_t) iio_be32toh(*(uint32_t *) ptr);
 		ptr += 4;
+		buf_size -= 4;
+
+		if (len > 0 && buf_size < len) {
+			ret = -EPROTO;
+			break;
+		}
+
 		if (len > 0) {
 			ret = cb(chn, attr, ptr, (size_t) len, data);
 			if (ret < 0)
@@ -766,6 +771,10 @@ int iio_channel_attr_read_all(struct iio_channel *chn,
 			if (len & 0x3)
 				len = ((len >> 2) + 1) << 2;
 			ptr += len;
+			if (len >= buf_size)
+				buf_size = 0;
+			else
+				buf_size -= len;
 		}
 	}
 
