@@ -1,20 +1,10 @@
+// SPDX-License-Identifier: LGPL-2.1-or-later
 /*
  * libiio - Library for interfacing industrial I/O (IIO) devices
  *
  * Copyright (C) 2014 Analog Devices, Inc.
  * Author: Paul Cercueil <paul.cercueil@analog.com>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * */
+ */
 
 #include "debug.h"
 #include "iio-private.h"
@@ -24,147 +14,115 @@
 #include <stdio.h>
 #include <string.h>
 
-static char *get_attr_xml(const char *attr, size_t *length, bool is_debug)
+static ssize_t iio_snprintf_xml_attr(char *buf, ssize_t len, const char *attr,
+				     enum iio_attr_type type)
 {
-	size_t len = sizeof("<attribute name=\"\" />") + strlen(attr)
-		+ (!is_debug ? 0 : sizeof("debug-") - 1);
-	char *str = malloc(len);
-	if (!str)
-		return NULL;
-
-	*length = len - 1; /* Skip the \0 */
-	if (is_debug)
-		iio_snprintf(str, len, "<debug-attribute name=\"%s\" />", attr);
-	else
-		iio_snprintf(str, len, "<attribute name=\"%s\" />", attr);
-	return str;
+	switch (type) {
+		case IIO_ATTR_TYPE_DEVICE:
+			return iio_snprintf(buf, len, "<attribute name=\"%s\" />", attr);
+		case IIO_ATTR_TYPE_DEBUG:
+			return iio_snprintf(buf, len, "<debug-attribute name=\"%s\" />", attr);
+		case IIO_ATTR_TYPE_BUFFER:
+			return iio_snprintf(buf, len, "<buffer-attribute name=\"%s\" />", attr);
+		default:
+			return -EINVAL;
+	}
 }
 
-/* Returns a string containing the XML representation of this device */
-char * iio_device_get_xml(const struct iio_device *dev, size_t *length)
+ssize_t iio_snprintf_device_xml(char *ptr, ssize_t len,
+				const struct iio_device *dev)
 {
-	size_t len = sizeof("<device id=\"\" name=\"\" ></device>")
-		+ strlen(dev->id) + (dev->name ? strlen(dev->name) : 0);
-	char *ptr, *str, **attrs, **channels, **debug_attrs;
-	size_t *attrs_len, *channels_len, *debug_attrs_len;
-	unsigned int i, j, k;
+	ssize_t ret, alen = 0;
+	unsigned int i;
 
-	attrs_len = malloc(dev->nb_attrs * sizeof(*attrs_len));
-	if (!attrs_len)
-		return NULL;
-
-	attrs = malloc(dev->nb_attrs * sizeof(*attrs));
-	if (!attrs)
-		goto err_free_attrs_len;
-
-	for (i = 0; i < dev->nb_attrs; i++) {
-		char *xml = get_attr_xml(dev->attrs[i], &attrs_len[i], false);
-		if (!xml)
-			goto err_free_attrs;
-		attrs[i] = xml;
-		len += attrs_len[i];
-	}
-
-	channels_len = malloc(dev->nb_channels * sizeof(*channels_len));
-	if (!channels_len)
-		goto err_free_attrs;
-
-	channels = malloc(dev->nb_channels * sizeof(*channels));
-	if (!channels)
-		goto err_free_channels_len;
-
-	for (j = 0; j < dev->nb_channels; j++) {
-		char *xml = iio_channel_get_xml(dev->channels[j],
-				&channels_len[j]);
-		if (!xml)
-			goto err_free_channels;
-		channels[j] = xml;
-		len += channels_len[j];
-	}
-
-	debug_attrs_len = malloc(dev->nb_debug_attrs *
-			sizeof(*debug_attrs_len));
-	if (!debug_attrs_len)
-		goto err_free_channels;
-
-	debug_attrs = malloc(dev->nb_debug_attrs * sizeof(*debug_attrs));
-	if (!debug_attrs)
-		goto err_free_debug_attrs_len;
-
-	for (k = 0; k < dev->nb_debug_attrs; k++) {
-		char *xml = get_attr_xml(dev->debug_attrs[k],
-				&debug_attrs_len[k], true);
-		if (!xml)
-			goto err_free_debug_attrs;
-		debug_attrs[k] = xml;
-		len += debug_attrs_len[k];
-	}
-
-	str = malloc(len);
-	if (!str)
-		goto err_free_debug_attrs;
-
-	iio_snprintf(str, len, "<device id=\"%s\"", dev->id);
-	ptr = strrchr(str, '\0');
+	ret = iio_snprintf(ptr, len, "<device id=\"%s\"", dev->id);
+	if (ret < 0)
+		return ret;
 
+	iio_update_xml_indexes(ret, &ptr, &len, &alen);
 	if (dev->name) {
-		sprintf(ptr, " name=\"%s\"", dev->name);
-		ptr = strrchr(ptr, '\0');
+		ret = iio_snprintf(ptr, len, " name=\"%s\"", dev->name);
+		if (ret < 0)
+			return ret;
+	
+		iio_update_xml_indexes(ret, &ptr, &len, &alen);
 	}
 
-	strcpy(ptr, " >");
-	ptr += 2;
+	if (dev->label) {
+		ret = iio_snprintf(ptr, len, " label=\"%s\"", dev->label);
+		if (ret < 0)
+			return ret;
+
+		iio_update_xml_indexes(ret, &ptr, &len, &alen);
+	}
+
+	ret = iio_snprintf(ptr, len, " >");
+	if (ret < 0)
+		return ret;
+	
+	iio_update_xml_indexes(ret, &ptr, &len, &alen);
 
 	for (i = 0; i < dev->nb_channels; i++) {
-		strcpy(ptr, channels[i]);
-		ptr += channels_len[i];
-		free(channels[i]);
+		ret = iio_snprintf_channel_xml(ptr, len, dev->channels[i]);
+		if (ret < 0)
+			return ret;
+
+		iio_update_xml_indexes(ret, &ptr, &len, &alen);
 	}
 
-	free(channels);
-	free(channels_len);
-
-	for (i = 0; i < dev->nb_attrs; i++) {
-		strcpy(ptr, attrs[i]);
-		ptr += attrs_len[i];
-		free(attrs[i]);
+	for (i = 0; i < dev->attrs.num; i++) {
+		ret = iio_snprintf_xml_attr(ptr, len, dev->attrs.names[i],
+					    IIO_ATTR_TYPE_DEVICE);
+		if (ret < 0)
+			return ret;
+		
+		iio_update_xml_indexes(ret, &ptr, &len, &alen);
 	}
 
-	free(attrs);
-	free(attrs_len);
+	for (i = 0; i < dev->buffer_attrs.num; i++) {
+		ret = iio_snprintf_xml_attr(ptr, len, dev->buffer_attrs.names[i],
+					    IIO_ATTR_TYPE_BUFFER);
+		if (ret < 0)
+			return ret;
 
-	for (i = 0; i < dev->nb_debug_attrs; i++) {
-		strcpy(ptr, debug_attrs[i]);
-		ptr += debug_attrs_len[i];
-		free(debug_attrs[i]);
+		iio_update_xml_indexes(ret, &ptr, &len, &alen);
 	}
 
-	free(debug_attrs);
-	free(debug_attrs_len);
+	for (i = 0; i < dev->debug_attrs.num; i++) {
+		ret = iio_snprintf_xml_attr(ptr, len, dev->debug_attrs.names[i],
+					    IIO_ATTR_TYPE_DEBUG);
+		if (ret < 0)
+			return ret;
 
-	strcpy(ptr, "</device>");
-	*length = ptr - str + sizeof("</device>") - 1;
-	return str;
+		iio_update_xml_indexes(ret, &ptr, &len, &alen);
+	}
 
-err_free_debug_attrs:
-	while (k--)
-		free(debug_attrs[k]);
-	free(debug_attrs);
-err_free_debug_attrs_len:
-	free(debug_attrs_len);
-err_free_channels:
-	while (j--)
-		free(channels[j]);
-	free(channels);
-err_free_channels_len:
-	free(channels_len);
-err_free_attrs:
-	while (i--)
-		free(attrs[i]);
-	free(attrs);
-err_free_attrs_len:
-	free(attrs_len);
-	return NULL;
+	ret = iio_snprintf(ptr, len, "</device>");
+	if (ret < 0)
+		return ret;
+
+	return alen + ret;
+}
+
+int add_iio_dev_attr(struct iio_dev_attrs *attrs, const char *attr,
+		     const char *type, const char *dev_id)
+{
+	char **names, *name;
+
+	name = iio_strdup(attr);
+	if (!name)
+		return -ENOMEM;
+
+	names = realloc(attrs->names, (1 + attrs->num) * sizeof(char *));
+	if (!names) {
+		free(name);
+		return -ENOMEM;
+	}
+
+	names[attrs->num++] = name;
+	attrs->names = names;
+	IIO_DEBUG("Added%s attr \'%s\' to device \'%s\'\n", type, attr, dev_id);
+	return 0;
 }
 
 const char * iio_device_get_id(const struct iio_device *dev)
@@ -177,6 +135,11 @@ const char * iio_device_get_name(const struct iio_device *dev)
 	return dev->name;
 }
 
+const char * iio_device_get_label(const struct iio_device *dev)
+{
+	return dev->label;
+}
+
 unsigned int iio_device_get_channels_count(const struct iio_device *dev)
 {
 	return dev->nb_channels;
@@ -207,42 +170,65 @@ struct iio_channel * iio_device_find_channel(const struct iio_device *dev,
 	return NULL;
 }
 
+static const char * iio_device_get_dev_attr(const struct iio_dev_attrs *attrs,
+		unsigned int index)
+{
+	if (index >= attrs->num)
+		return NULL;
+	else
+		return attrs->names[index];
+}
+
+const char * iio_device_find_dev_attr(const struct iio_dev_attrs *attrs,
+		const char *name)
+{
+	unsigned int i;
+	for (i = 0; i < attrs->num; i++) {
+		const char *attr = attrs->names[i];
+		if (!strcmp(attr, name))
+			return attr;
+	}
+	return NULL;
+}
+
 unsigned int iio_device_get_attrs_count(const struct iio_device *dev)
 {
-	return dev->nb_attrs;
+	return dev->attrs.num;
 }
 
 const char * iio_device_get_attr(const struct iio_device *dev,
 		unsigned int index)
 {
-	if (index >= dev->nb_attrs)
-		return NULL;
-	else
-		return dev->attrs[index];
+	return iio_device_get_dev_attr(&dev->attrs, index);
 }
 
 const char * iio_device_find_attr(const struct iio_device *dev,
 		const char *name)
 {
-	unsigned int i;
-	for (i = 0; i < dev->nb_attrs; i++) {
-		const char *attr = dev->attrs[i];
-		if (!strcmp(attr, name))
-			return attr;
-	}
-	return NULL;
+	return iio_device_find_dev_attr(&dev->attrs, name);
+}
+
+unsigned int iio_device_get_buffer_attrs_count(const struct iio_device *dev)
+{
+	return dev->buffer_attrs.num;
+}
+
+const char * iio_device_get_buffer_attr(const struct iio_device *dev,
+		unsigned int index)
+{
+	return iio_device_get_dev_attr(&dev->buffer_attrs, index);
+}
+
+const char * iio_device_find_buffer_attr(const struct iio_device *dev,
+		const char *name)
+{
+	return iio_device_find_dev_attr(&dev->buffer_attrs, name);
 }
 
 const char * iio_device_find_debug_attr(const struct iio_device *dev,
 		const char *name)
 {
-	unsigned int i;
-	for (i = 0; i < dev->nb_debug_attrs; i++) {
-		const char *attr = dev->debug_attrs[i];
-		if (!strcmp(attr, name))
-			return attr;
-	}
-	return NULL;
+	return iio_device_find_dev_attr(&dev->debug_attrs, name);
 }
 
 bool iio_device_is_tx(const struct iio_device *dev)
@@ -322,7 +308,7 @@ ssize_t iio_device_attr_read(const struct iio_device *dev,
 {
 	if (dev->ctx->ops->read_device_attr)
 		return dev->ctx->ops->read_device_attr(dev,
-				attr, dst, len, false);
+				attr, dst, len, IIO_ATTR_TYPE_DEVICE);
 	else
 		return -ENOSYS;
 }
@@ -332,7 +318,7 @@ ssize_t iio_device_attr_write_raw(const struct iio_device *dev,
 {
 	if (dev->ctx->ops->write_device_attr)
 		return dev->ctx->ops->write_device_attr(dev,
-				attr, src, len, false);
+				attr, src, len, IIO_ATTR_TYPE_DEVICE);
 	else
 		return -ENOSYS;
 }
@@ -343,6 +329,32 @@ ssize_t iio_device_attr_write(const struct iio_device *dev,
 	return iio_device_attr_write_raw(dev, attr, src, strlen(src) + 1);
 }
 
+ssize_t iio_device_buffer_attr_read(const struct iio_device *dev,
+		const char *attr, char *dst, size_t len)
+{
+	if (dev->ctx->ops->read_device_attr)
+		return dev->ctx->ops->read_device_attr(dev,
+				attr, dst, len, IIO_ATTR_TYPE_BUFFER);
+	else
+		return -ENOSYS;
+}
+
+ssize_t iio_device_buffer_attr_write_raw(const struct iio_device *dev,
+		const char *attr, const void *src, size_t len)
+{
+	if (dev->ctx->ops->write_device_attr)
+		return dev->ctx->ops->write_device_attr(dev,
+				attr, src, len, IIO_ATTR_TYPE_BUFFER);
+	else
+		return -ENOSYS;
+}
+
+ssize_t iio_device_buffer_attr_write(const struct iio_device *dev,
+		const char *attr, const char *src)
+{
+	return iio_device_buffer_attr_write_raw(dev, attr, src, strlen(src) + 1);
+}
+
 void iio_device_set_data(struct iio_device *dev, void *data)
 {
 	dev->userdata = data;
@@ -398,27 +410,31 @@ int iio_device_set_trigger(const struct iio_device *dev,
 		return -ENOSYS;
 }
 
+static void free_iio_dev_attrs(struct iio_dev_attrs *attrs)
+{
+	unsigned int i;
+
+	for (i = 0; i <  attrs->num; i++)
+		free(attrs->names[i]);
+
+	free(attrs->names);
+}
+
 void free_device(struct iio_device *dev)
 {
 	unsigned int i;
-	for (i = 0; i < dev->nb_attrs; i++)
-		free(dev->attrs[i]);
-	if (dev->nb_attrs)
-		free(dev->attrs);
-	for (i = 0; i < dev->nb_debug_attrs; i++)
-		free(dev->debug_attrs[i]);
-	if (dev->nb_debug_attrs)
-		free(dev->debug_attrs);
+
+	free_iio_dev_attrs(&dev->attrs);
+	free_iio_dev_attrs(&dev->buffer_attrs);
+	free_iio_dev_attrs(&dev->debug_attrs);
+
 	for (i = 0; i < dev->nb_channels; i++)
 		free_channel(dev->channels[i]);
-	if (dev->nb_channels)
-		free(dev->channels);
-	if (dev->mask)
-		free(dev->mask);
-	if (dev->name)
-		free(dev->name);
-	if (dev->id)
-		free(dev->id);
+	free(dev->channels);
+	free(dev->mask);
+	free(dev->label);
+	free(dev->name);
+	free(dev->id);
 	free(dev);
 }
 
@@ -471,8 +487,9 @@ int iio_device_attr_read_longlong(const struct iio_device *dev,
 	if (ret < 0)
 		return (int) ret;
 
+	errno = 0;
 	value = strtoll(buf, &end, 0);
-	if (end == buf)
+	if (end == buf || errno == ERANGE)
 		return -EINVAL;
 	*val = value;
 	return 0;
@@ -510,7 +527,7 @@ int iio_device_attr_write_longlong(const struct iio_device *dev,
 	iio_snprintf(buf, sizeof(buf), "%lld", val);
 	ret = iio_device_attr_write(dev, attr, buf);
 
-	return ret < 0 ? ret : 0;
+	return (int) (ret < 0 ? ret : 0);
 }
 
 int iio_device_attr_write_double(const struct iio_device *dev,
@@ -522,7 +539,7 @@ int iio_device_attr_write_double(const struct iio_device *dev,
 	ret = (ssize_t) write_double(buf, sizeof(buf), val);
 	if (!ret)
 		ret = iio_device_attr_write(dev, attr, buf);
-	return ret < 0 ? ret : 0;
+	return (int) (ret < 0 ? ret : 0);
 }
 
 int iio_device_attr_write_bool(const struct iio_device *dev,
@@ -535,7 +552,84 @@ int iio_device_attr_write_bool(const struct iio_device *dev,
 	else
 		ret = iio_device_attr_write(dev, attr, "0");
 
-	return ret < 0 ? ret : 0;
+	return (int) (ret < 0 ? ret : 0);
+}
+
+int iio_device_buffer_attr_read_longlong(const struct iio_device *dev,
+		const char *attr, long long *val)
+{
+	char *end, buf[1024];
+	long long value;
+	ssize_t ret = iio_device_buffer_attr_read(dev, attr, buf, sizeof(buf));
+	if (ret < 0)
+		return (int) ret;
+
+	errno = 0;
+	value = strtoll(buf, &end, 0);
+	if (end == buf || errno == ERANGE)
+		return -EINVAL;
+	*val = value;
+	return 0;
+}
+
+int iio_device_buffer_attr_read_bool(const struct iio_device *dev,
+		const char *attr, bool *val)
+{
+	long long value;
+	int ret = iio_device_buffer_attr_read_longlong(dev, attr, &value);
+	if (ret < 0)
+		return ret;
+
+	*val = !!value;
+	return 0;
+}
+
+int iio_device_buffer_attr_read_double(const struct iio_device *dev,
+		const char *attr, double *val)
+{
+	char buf[1024];
+	ssize_t ret = iio_device_buffer_attr_read(dev, attr, buf, sizeof(buf));
+	if (ret < 0)
+		return (int) ret;
+	else
+		return read_double(buf, val);
+}
+
+int iio_device_buffer_attr_write_longlong(const struct iio_device *dev,
+		const char *attr, long long val)
+{
+	ssize_t ret;
+	char buf[1024];
+
+	iio_snprintf(buf, sizeof(buf), "%lld", val);
+	ret = iio_device_buffer_attr_write(dev, attr, buf);
+
+	return (int) (ret < 0 ? ret : 0);
+}
+
+int iio_device_buffer_attr_write_double(const struct iio_device *dev,
+		const char *attr, double val)
+{
+	ssize_t ret;
+	char buf[1024];
+
+	ret = (ssize_t) write_double(buf, sizeof(buf), val);
+	if (!ret)
+		ret = iio_device_buffer_attr_write(dev, attr, buf);
+	return (int) (ret < 0 ? ret : 0);
+}
+
+int iio_device_buffer_attr_write_bool(const struct iio_device *dev,
+		const char *attr, bool val)
+{
+	ssize_t ret;
+
+	if (val)
+		ret = iio_device_buffer_attr_write(dev, attr, "1");
+	else
+		ret = iio_device_buffer_attr_write(dev, attr, "0");
+
+	return (int) (ret < 0 ? ret : 0);
 }
 
 ssize_t iio_device_debug_attr_read(const struct iio_device *dev,
@@ -543,7 +637,7 @@ ssize_t iio_device_debug_attr_read(const struct iio_device *dev,
 {
 	if (dev->ctx->ops->read_device_attr)
 		return dev->ctx->ops->read_device_attr(dev,
-				attr, dst, len, true);
+				attr, dst, len, IIO_ATTR_TYPE_DEBUG);
 	else
 		return -ENOSYS;
 }
@@ -553,7 +647,7 @@ ssize_t iio_device_debug_attr_write_raw(const struct iio_device *dev,
 {
 	if (dev->ctx->ops->write_device_attr)
 		return dev->ctx->ops->write_device_attr(dev,
-				attr, src, len, true);
+				attr, src, len, IIO_ATTR_TYPE_DEBUG);
 	else
 		return -ENOSYS;
 }
@@ -566,16 +660,13 @@ ssize_t iio_device_debug_attr_write(const struct iio_device *dev,
 
 unsigned int iio_device_get_debug_attrs_count(const struct iio_device *dev)
 {
-	return dev->nb_debug_attrs;
+	return dev->debug_attrs.num;
 }
 
 const char * iio_device_get_debug_attr(const struct iio_device *dev,
 		unsigned int index)
 {
-	if (index >= dev->nb_debug_attrs)
-		return NULL;
-	else
-		return dev->debug_attrs[index];
+	return iio_device_get_dev_attr(&dev->debug_attrs, index);
 }
 
 int iio_device_debug_attr_read_longlong(const struct iio_device *dev,
@@ -587,8 +678,9 @@ int iio_device_debug_attr_read_longlong(const struct iio_device *dev,
 	if (ret < 0)
 		return (int) ret;
 
+	errno = 0;
 	value = strtoll(buf, &end, 0);
-	if (end == buf)
+	if (end == buf || errno == ERANGE)
 		return -EINVAL;
 	*val = value;
 	return 0;
@@ -626,7 +718,7 @@ int iio_device_debug_attr_write_longlong(const struct iio_device *dev,
 	iio_snprintf(buf, sizeof(buf), "%lld", val);
 	ret = iio_device_debug_attr_write(dev, attr, buf);
 
-	return ret < 0 ? ret : 0;
+	return (int) (ret < 0 ? ret : 0);
 }
 
 int iio_device_debug_attr_write_double(const struct iio_device *dev,
@@ -638,7 +730,7 @@ int iio_device_debug_attr_write_double(const struct iio_device *dev,
 	ret = (ssize_t) write_double(buf, sizeof(buf), val);
 	if (!ret)
 		ret = iio_device_debug_attr_write(dev, attr, buf);
-	return ret < 0 ? ret : 0;
+	return (int) (ret < 0 ? ret : 0);
 }
 
 int iio_device_debug_attr_write_bool(const struct iio_device *dev,
@@ -651,7 +743,7 @@ int iio_device_debug_attr_write_bool(const struct iio_device *dev,
 	else
 		ret = iio_device_debug_attr_write_raw(dev, attr, "0", 2);
 
-	return ret < 0 ? ret : 0;
+	return (int) (ret < 0 ? ret : 0);
 }
 
 int iio_device_identify_filename(const struct iio_device *dev,
@@ -673,18 +765,18 @@ int iio_device_identify_filename(const struct iio_device *dev,
 		}
 	}
 
-	for (i = 0; i < dev->nb_attrs; i++) {
+	for (i = 0; i < dev->attrs.num; i++) {
 		/* Devices attributes are named after their filename */
-		if (!strcmp(dev->attrs[i], filename)) {
-			*attr = dev->attrs[i];
+		if (!strcmp(dev->attrs.names[i], filename)) {
+			*attr = dev->attrs.names[i];
 			*chn = NULL;
 			return 0;
 		}
 	}
 
-	for (i = 0; i < dev->nb_debug_attrs; i++) {
-		if (!strcmp(dev->debug_attrs[i], filename)) {
-			*attr = dev->debug_attrs[i];
+	for (i = 0; i < dev->debug_attrs.num; i++) {
+		if (!strcmp(dev->debug_attrs.names[i], filename)) {
+			*attr = dev->debug_attrs.names[i];
 			*chn = NULL;
 			return 0;
 		}
@@ -703,7 +795,7 @@ int iio_device_reg_write(struct iio_device *dev,
 			address, value);
 	ret = iio_device_debug_attr_write(dev, "direct_reg_access", buf);
 
-	return ret < 0 ? ret : 0;
+	return (int) (ret < 0 ? ret : 0);
 }
 
 int iio_device_reg_read(struct iio_device *dev,
@@ -726,12 +818,12 @@ int iio_device_reg_read(struct iio_device *dev,
 	return ret;
 }
 
-static int read_each_attr(struct iio_device *dev, bool is_debug,
+static int read_each_attr(struct iio_device *dev, enum iio_attr_type type,
 		int (*cb)(struct iio_device *dev,
 			const char *attr, const char *val, size_t len, void *d),
 		void *data)
 {
-	int ret;
+	int ret, buf_size;
 	char *buf, *ptr;
 	unsigned int i, count;
 
@@ -740,30 +832,67 @@ static int read_each_attr(struct iio_device *dev, bool is_debug,
 	if (!buf)
 		return -ENOMEM;
 
-	if (is_debug) {
-		count = iio_device_get_debug_attrs_count(dev);
-		ret = (int) iio_device_debug_attr_read(dev,
-				NULL, buf, 0x100000);
-	} else {
-		count = iio_device_get_attrs_count(dev);
-		ret = (int) iio_device_attr_read(dev, NULL, buf, 0x100000);
+	switch(type){
+		case IIO_ATTR_TYPE_DEVICE:
+			count = iio_device_get_attrs_count(dev);
+			ret = (int) iio_device_attr_read(dev,
+					NULL, buf, 0x100000);
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			count = iio_device_get_debug_attrs_count(dev);
+			ret = (int) iio_device_debug_attr_read(dev,
+					NULL, buf, 0x100000);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			count = iio_device_get_buffer_attrs_count(dev);
+			ret = (int) iio_device_buffer_attr_read(dev,
+					NULL, buf, 0x100000);
+			break;
+		default:
+			ret = -EINVAL;
+			count = 0;
+			break;
 	}
 
 	if (ret < 0)
 		goto err_free_buf;
 
 	ptr = buf;
+	buf_size = ret;
 
 	for (i = 0; i < count; i++) {
 		const char *attr;
-		int32_t len = (int32_t) iio_be32toh(*(uint32_t *) ptr);
+		int32_t len;
 
-		if (is_debug)
-			attr = iio_device_get_debug_attr(dev, i);
-		else
-			attr = iio_device_get_attr(dev, i);
+		if (buf_size < 4) {
+			ret = -EPROTO;
+			break;
+		}
 
+		len = (int32_t) iio_be32toh(*(uint32_t *) ptr);
 		ptr += 4;
+		buf_size -= 4;
+
+		if (len > 0 && buf_size < len) {
+			ret = -EPROTO;
+			break;
+		}
+
+		switch(type){
+			case IIO_ATTR_TYPE_DEVICE:
+				attr = iio_device_get_attr(dev, i);
+				break;
+			case IIO_ATTR_TYPE_DEBUG:
+				attr = iio_device_get_debug_attr(dev, i);
+				break;
+			case IIO_ATTR_TYPE_BUFFER:
+				attr = iio_device_get_buffer_attr(dev, i);
+				break;
+			default:
+				attr = NULL;
+				break;
+		}
+
 		if (len > 0) {
 			ret = cb(dev, attr, ptr, (size_t) len, data);
 			if (ret < 0)
@@ -772,6 +901,10 @@ static int read_each_attr(struct iio_device *dev, bool is_debug,
 			if (len & 0x3)
 				len = ((len >> 2) + 1) << 2;
 			ptr += len;
+			if (len >= buf_size)
+				buf_size = 0;
+			else
+				buf_size -= len;
 		}
 	}
 
@@ -780,7 +913,7 @@ err_free_buf:
 	return ret < 0 ? ret : 0;
 }
 
-static int write_each_attr(struct iio_device *dev, bool is_debug,
+static int write_each_attr(struct iio_device *dev, enum iio_attr_type type,
 		ssize_t (*cb)(struct iio_device *dev,
 			const char *attr, void *buf, size_t len, void *d),
 		void *data)
@@ -797,18 +930,38 @@ static int write_each_attr(struct iio_device *dev, bool is_debug,
 
 	ptr = buf;
 
-	if (is_debug)
-		count = iio_device_get_debug_attrs_count(dev);
-	else
-		count = iio_device_get_attrs_count(dev);
+	switch(type){
+		case IIO_ATTR_TYPE_DEVICE:
+			count = iio_device_get_attrs_count(dev);
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			count = iio_device_get_debug_attrs_count(dev);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			count = iio_device_get_buffer_attrs_count(dev);
+			break;
+		default:
+			ret = -EINVAL;
+			goto err_free_buf;
+	}
 
 	for (i = 0; i < count; i++) {
 		const char *attr;
 
-		if (is_debug)
-			attr = iio_device_get_debug_attr(dev, i);
-		else
-			attr = iio_device_get_attr(dev, i);
+		switch(type){
+			case IIO_ATTR_TYPE_DEVICE:
+				attr = iio_device_get_attr(dev, i);
+				break;
+			case IIO_ATTR_TYPE_DEBUG:
+				attr = iio_device_get_debug_attr(dev, i);
+				break;
+			case IIO_ATTR_TYPE_BUFFER:
+				attr = iio_device_get_buffer_attr(dev, i);
+				break;
+			default:
+				attr = NULL;
+				break;
+		}
 
 		ret = (int) cb(dev, attr, ptr + 4, len - 4, data);
 		if (ret < 0)
@@ -826,12 +979,23 @@ static int write_each_attr(struct iio_device *dev, bool is_debug,
 		}
 	}
 
-	if (is_debug)
-		ret = (int) iio_device_debug_attr_write_raw(dev,
-				NULL, buf, ptr - buf);
-	else
-		ret = (int) iio_device_attr_write_raw(dev,
-				NULL, buf, ptr - buf);
+	switch(type){
+		case IIO_ATTR_TYPE_DEVICE:
+			ret = (int) iio_device_attr_write_raw(dev,
+					NULL, buf, ptr - buf);
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			ret = (int) iio_device_debug_attr_write_raw(dev,
+					NULL, buf, ptr - buf);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			ret = (int) iio_device_buffer_attr_write_raw(dev,
+					NULL, buf, ptr - buf);
+			break;
+		default:
+			ret = -EINVAL;
+			break;
+	}
 
 err_free_buf:
 	free(buf);
@@ -843,7 +1007,15 @@ int iio_device_debug_attr_read_all(struct iio_device *dev,
 			const char *attr, const char *val, size_t len, void *d),
 		void *data)
 {
-	return read_each_attr(dev, true, cb, data);
+	return read_each_attr(dev, IIO_ATTR_TYPE_DEBUG, cb, data);
+}
+
+int iio_device_buffer_attr_read_all(struct iio_device *dev,
+		int (*cb)(struct iio_device *dev,
+			const char *attr, const char *val, size_t len, void *d),
+		void *data)
+{
+	return read_each_attr(dev, IIO_ATTR_TYPE_BUFFER, cb, data);
 }
 
 int iio_device_attr_read_all(struct iio_device *dev,
@@ -851,7 +1023,7 @@ int iio_device_attr_read_all(struct iio_device *dev,
 			const char *attr, const char *val, size_t len, void *d),
 		void *data)
 {
-	return read_each_attr(dev, false, cb, data);
+	return read_each_attr(dev, IIO_ATTR_TYPE_DEVICE ,cb, data);
 }
 
 int iio_device_debug_attr_write_all(struct iio_device *dev,
@@ -859,7 +1031,15 @@ int iio_device_debug_attr_write_all(struct iio_device *dev,
 			const char *attr, void *buf, size_t len, void *d),
 		void *data)
 {
-	return write_each_attr(dev, true, cb, data);
+	return write_each_attr(dev, IIO_ATTR_TYPE_DEBUG, cb, data);
+}
+
+int iio_device_buffer_attr_write_all(struct iio_device *dev,
+		ssize_t (*cb)(struct iio_device *dev,
+			const char *attr, void *buf, size_t len, void *d),
+		void *data)
+{
+	return write_each_attr(dev, IIO_ATTR_TYPE_BUFFER, cb, data);
 }
 
 int iio_device_attr_write_all(struct iio_device *dev,
@@ -867,7 +1047,7 @@ int iio_device_attr_write_all(struct iio_device *dev,
 			const char *attr, void *buf, size_t len, void *d),
 		void *data)
 {
-	return write_each_attr(dev, false, cb, data);
+	return write_each_attr(dev, IIO_ATTR_TYPE_DEVICE, cb, data);
 }
 
 const struct iio_context * iio_device_get_context(const struct iio_device *dev)
