@@ -1,26 +1,15 @@
+// SPDX-License-Identifier: LGPL-2.1-or-later
 /*
  * libiio - Library for interfacing industrial I/O (IIO) devices
  *
  * Copyright (C) 2014 Analog Devices, Inc.
  * Author: Paul Cercueil <paul.cercueil@analog.com>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * */
+ */
 
 #include "ops.h"
 #include "parser.h"
 #include "thread-pool.h"
 #include "../debug.h"
-#include "../iio-private.h"
 
 #include <errno.h>
 #include <limits.h>
@@ -105,6 +94,7 @@ struct DevEntry {
 	struct iio_device *dev;
 	struct iio_buffer *buf;
 	unsigned int sample_size, nb_clients;
+	unsigned int samples_count;
 	bool update_mask;
 	bool cyclic;
 	bool closed;
@@ -131,6 +121,36 @@ struct sample_cb_info {
  * clients */
 static pthread_mutex_t devlist_lock = PTHREAD_MUTEX_INITIALIZER;
 
+static unsigned int get_channel_number(const struct iio_channel *chn)
+{
+	const struct iio_device *dev = iio_channel_get_device(chn);
+	const struct iio_channel *other;
+	unsigned int i = 0;
+
+	for (i = 0; i < iio_device_get_channels_count(dev); i++) {
+		other = iio_device_get_channel(dev, i);
+		if (other == chn)
+			break;
+	}
+
+	return i;
+}
+
+static inline const char *dev_label_or_name_or_id(const struct iio_device *dev)
+{
+	const char *name;
+
+	name = iio_device_get_label(dev);
+	if (name)
+		return name;
+
+	name = iio_device_get_name(dev);
+	if (name)
+		return name;
+
+	return iio_device_get_id(dev);
+}
+
 #if WITH_AIO
 static ssize_t async_io(struct parser_pdata *pdata, void *buf, size_t len,
 	bool do_read)
@@ -156,7 +176,7 @@ static ssize_t async_io(struct parser_pdata *pdata, void *buf, size_t len,
 	ret = io_submit(pdata->aio_ctx, 1, ios);
 	if (ret != 1) {
 		pthread_mutex_unlock(&pdata->aio_mutex);
-		ERROR("Failed to submit IO operation: %zd\n", ret);
+		IIO_ERROR("Failed to submit IO operation: %zd\n", ret);
 		return -EIO;
 	}
 
@@ -175,14 +195,14 @@ static ssize_t async_io(struct parser_pdata *pdata, void *buf, size_t len,
 			uint64_t event;
 			ret = read(pdata->aio_eventfd, &event, sizeof(event));
 			if (ret != sizeof(event)) {
-				ERROR("Failed to read from eventfd: %d\n", -errno);
+				IIO_ERROR("Failed to read from eventfd: %d\n", -errno);
 				ret = -EIO;
 				break;
 			}
 
 			ret = io_getevents(pdata->aio_ctx, 0, 1, e, NULL);
 			if (ret != 1) {
-				ERROR("Failed to read IO events: %zd\n", ret);
+				IIO_ERROR("Failed to read IO events: %zd\n", ret);
 				ret = -EIO;
 				break;
 			} else {
@@ -192,7 +212,7 @@ static ssize_t async_io(struct parser_pdata *pdata, void *buf, size_t len,
 			/* Got a STOP event to abort this whole session */
 			ret = io_cancel(pdata->aio_ctx, &iocb, e);
 			if (ret != -EINPROGRESS && ret != -EINVAL) {
-				ERROR("Failed to cancel IO transfer: %zd\n", ret);
+				IIO_ERROR("Failed to cancel IO transfer: %zd\n", ret);
 				ret = -EIO;
 				break;
 			}
@@ -353,7 +373,7 @@ static void print_value(struct parser_pdata *pdata, long value)
 		output(pdata, "\n");
 	} else {
 		char buf[128];
-		sprintf(buf, "%li\n", value);
+		snprintf(buf, sizeof(buf), "%li\n", value);
 		output(pdata, buf);
 	}
 }
@@ -361,8 +381,10 @@ static void print_value(struct parser_pdata *pdata, long value)
 static ssize_t send_sample(const struct iio_channel *chn,
 		void *src, size_t length, void *d)
 {
+	unsigned int number = get_channel_number(chn);
 	struct sample_cb_info *info = d;
-	if (chn->index < 0 || !TEST_BIT(info->mask, chn->index))
+
+	if (iio_channel_get_index(chn) < 0 || !TEST_BIT(info->mask, number))
 		return 0;
 	if (info->nb_bytes < length)
 		return 0;
@@ -388,8 +410,10 @@ static ssize_t send_sample(const struct iio_channel *chn,
 static ssize_t receive_sample(const struct iio_channel *chn,
 		void *dst, size_t length, void *d)
 {
+	unsigned int number = get_channel_number(chn);
 	struct sample_cb_info *info = d;
-	if (chn->index < 0 || !TEST_BIT(info->mask, chn->index))
+
+	if (iio_channel_get_index(chn) < 0 || !TEST_BIT(info->mask, number))
 		return 0;
 	if (info->cpt == info->nb_bytes)
 		return 0;
@@ -416,6 +440,7 @@ static ssize_t send_data(struct DevEntry *dev, struct ThdEntry *thd, size_t len)
 {
 	struct parser_pdata *pdata = thd->pdata;
 	bool demux = server_demux && dev->sample_size != thd->sample_size;
+	void *start;
 
 	if (demux)
 		len = (len / dev->sample_size) * thd->sample_size;
@@ -428,14 +453,24 @@ static ssize_t send_data(struct DevEntry *dev, struct ThdEntry *thd, size_t len)
 		unsigned int i;
 		char buf[129], *ptr = buf;
 		uint32_t *mask = demux ? thd->mask : dev->mask;
-		ssize_t ret;
+		ssize_t ret, length;
 
+		length = sizeof(buf);
 		/* Send the current mask */
 		for (i = dev->nb_words; i > 0 && ptr < buf + sizeof(buf);
-				i--, ptr += 8)
-			sprintf(ptr, "%08x", mask[i - 1]);
+				i--, ptr += 8) {
+			snprintf(ptr, length, "%08x", mask[i - 1]);
+			length -= 8;
+		}
 
 		*ptr = '\n';
+		length--;
+
+		if (length < 0) {
+			IIO_ERROR("send_data: string length error\n");
+			return -ENOSPC;
+		}
+
 		ret = write_all(pdata, buf, ptr + 1 - buf);
 		if (ret < 0)
 			return ret;
@@ -445,7 +480,8 @@ static ssize_t send_data(struct DevEntry *dev, struct ThdEntry *thd, size_t len)
 
 	if (!demux) {
 		/* Short path */
-		return write_all(pdata, dev->buf->buffer, len);
+		start = iio_buffer_start(dev->buf);
+		return write_all(pdata, start, len);
 	} else {
 		struct sample_cb_info info = {
 			.pdata = pdata,
@@ -462,7 +498,7 @@ static ssize_t receive_data(struct DevEntry *dev, struct ThdEntry *thd)
 {
 	struct parser_pdata *pdata = thd->pdata;
 
-	/* Inform that no error occured, and that we'll start reading data */
+	/* Inform that no error occurred, and that we'll start reading data */
 	if (thd->new_client) {
 		print_value(thd->pdata, 0);
 		thd->new_client = false;
@@ -471,11 +507,11 @@ static ssize_t receive_data(struct DevEntry *dev, struct ThdEntry *thd)
 	if (dev->sample_size == thd->sample_size) {
 		/* Short path: Receive directly in the buffer */
 
-		size_t len = dev->buf->length;
+		size_t len = dev->sample_size * dev->samples_count;
 		if (thd->nb < len)
 			len = thd->nb;
 
-		return read_all(pdata, dev->buf->buffer, len);
+		return read_all(pdata, iio_buffer_start(dev->buf), len);
 	} else {
 		/* Long path: Mux the samples to the buffer */
 
@@ -526,8 +562,8 @@ static void rw_thd(struct thread_pool *pool, void *d)
 	unsigned int nb_words = entry->nb_words;
 	ssize_t ret = 0;
 
-	DEBUG("R/W thread started for device %s\n",
-			dev->name ? dev->name : dev->id);
+	IIO_DEBUG("R/W thread started for device %s\n",
+		  dev_label_or_name_or_id(dev));
 
 	while (true) {
 		bool has_readers = false, has_writers = false,
@@ -556,11 +592,15 @@ static void rw_thd(struct thread_pool *pool, void *d)
 			if (entry->buf)
 				iio_buffer_destroy(entry->buf);
 
-			for (i = 0; i < dev->nb_channels; i++) {
-				struct iio_channel *chn = dev->channels[i];
-				long index = chn->index;
+			for (i = 0; i < iio_device_get_channels_count(dev); i++) {
+				struct iio_channel *chn = iio_device_get_channel(dev, i);
+				unsigned int number = get_channel_number(chn);
+				long index = iio_channel_get_index(chn);
 
-				if (index >= 0 && TEST_BIT(entry->mask, i))
+				if (index < 0)
+					continue;
+
+				if (TEST_BIT(entry->mask, number))
 					iio_channel_enable(chn);
 				else
 					iio_channel_disable(chn);
@@ -570,7 +610,7 @@ static void rw_thd(struct thread_pool *pool, void *d)
 					samples_count, entry->cyclic);
 			if (!entry->buf) {
 				ret = -errno;
-				ERROR("Unable to create buffer\n");
+				IIO_ERROR("Unable to create buffer\n");
 				break;
 			}
 			entry->cancelled = false;
@@ -583,13 +623,14 @@ static void rw_thd(struct thread_pool *pool, void *d)
 				}
 			}
 
-			DEBUG("IIO device %s reopened with new mask:\n",
-					dev->id);
+			IIO_DEBUG("IIO device %s reopened with new mask:\n",
+				  dev_label_or_name_or_id(dev));
 			for (i = 0; i < nb_words; i++)
-				DEBUG("Mask[%i] = 0x%08x\n", i, entry->mask[i]);
+				IIO_DEBUG("Mask[%i] = 0x%08x\n", i, entry->mask[i]);
 			entry->update_mask = false;
 
 			entry->sample_size = iio_device_get_sample_size(dev);
+			entry->samples_count = samples_count;
 			mask_updated = true;
 		}
 
@@ -637,7 +678,7 @@ static void rw_thd(struct thread_pool *pool, void *d)
 			}
 
 			if (ret < 0) {
-				ERROR("Reading from device failed: %i\n",
+				IIO_ERROR("Reading from device failed: %i\n",
 						(int) ret);
 				break;
 			}
@@ -661,7 +702,7 @@ static void rw_thd(struct thread_pool *pool, void *d)
 
 				if (ret < 0 || thd->nb < sample_size)
 					signal_thread(thd, (ret < 0) ?
-							ret : thd->nb);
+							ret : (ssize_t) thd->nb);
 			}
 
 			pthread_mutex_unlock(&entry->thdlist_lock);
@@ -672,8 +713,17 @@ static void rw_thd(struct thread_pool *pool, void *d)
 
 			pthread_mutex_lock(&entry->thdlist_lock);
 
-			/* Reset the size of the buffer to its maximum size */
-			entry->buf->data_length = entry->buf->length;
+			/* Reset the size of the buffer to its maximum size.
+			 *
+			 * XXX(pcercuei): There is no way to perform this with
+			 * the public libiio API. However, it probably does not
+			 * matter; we only need to reset the size of the buffer
+			 * if the buffer was used for receiving samples, and
+			 * to date there is no IIO device that supports both
+			 * receiving and sending samples.
+			 *
+			 * entry->buf->data_length = entry->buf->length;
+			 */
 
 			/* Same comment as above */
 			for (thd = SLIST_FIRST(&entry->thdlist_head);
@@ -701,7 +751,7 @@ static void rw_thd(struct thread_pool *pool, void *d)
 				continue;
 			}
 			if (ret < 0) {
-				ERROR("Writing to device failed: %i\n",
+				IIO_ERROR("Writing to device failed: %i\n",
 						(int) ret);
 				break;
 			}
@@ -740,8 +790,8 @@ static void rw_thd(struct thread_pool *pool, void *d)
 		iio_device_set_data(dev, NULL);
 	pthread_mutex_unlock(&devlist_lock);
 
-	DEBUG("Stopping R/W thread for device %s\n",
-			dev->name ? dev->name : dev->id);
+	IIO_DEBUG("Stopping R/W thread for device %s\n",
+		  dev_label_or_name_or_id(dev));
 
 	dev_entry_put(entry);
 }
@@ -797,7 +847,7 @@ static ssize_t rw_buffer(struct parser_pdata *pdata,
 
 	pthread_cond_signal(&entry->rw_ready_cond);
 
-	DEBUG("Waiting for completion...\n");
+	IIO_DEBUG("Waiting for completion...\n");
 	while (thd->active) {
 		ret = thd_entry_event_wait(thd, &entry->thdlist_lock, pdata->fd_in);
 		if (ret)
@@ -807,10 +857,10 @@ static ssize_t rw_buffer(struct parser_pdata *pdata,
 		ret = thd->err;
 	pthread_mutex_unlock(&entry->thdlist_lock);
 
-	if (ret > 0 && ret < nb)
+	if (ret > 0 && ret < (ssize_t) nb)
 		print_value(thd->pdata, 0);
 
-	DEBUG("Exiting rw_buffer with code %li\n", (long) ret);
+	IIO_DEBUG("Exiting rw_buffer with code %li\n", (long) ret);
 	if (ret < 0)
 		return ret;
 	else
@@ -827,10 +877,10 @@ static uint32_t *get_mask(const char *mask, size_t *len)
 	ptr = words + nb;
 	while (*mask) {
 		char buf[9];
-		sprintf(buf, "%.*s", 8, mask);
+		snprintf(buf, sizeof(buf), "%.*s", 8, mask);
 		sscanf(buf, "%08x", --ptr);
 		mask += 8;
-		DEBUG("Mask[%lu]: 0x%08x\n",
+		IIO_DEBUG("Mask[%lu]: 0x%08x\n",
 				(unsigned long) (words - ptr) / 4, *ptr);
 	}
 
@@ -866,6 +916,48 @@ static void remove_thd_entry(struct ThdEntry *t)
 	free_thd_entry(t);
 }
 
+static ssize_t get_dev_sample_size_mask(const struct iio_device *dev,
+					const uint32_t *mask, size_t words)
+{
+	unsigned int i, len, number,
+		     nb_channels = iio_device_get_channels_count(dev);
+	const struct iio_channel *prev = NULL;
+	const struct iio_channel *chn;
+	const struct iio_data_format *fmt;
+	long index;
+	ssize_t size = 0;
+
+	if (words != (nb_channels + 31) / 32)
+		return -EINVAL;
+
+	for (i = 0; i < nb_channels; i++) {
+		chn = iio_device_get_channel(dev, i);
+		number = get_channel_number(chn);
+		fmt = iio_channel_get_data_format(chn);
+		index = iio_channel_get_index(chn);
+		len = fmt->length / 8 * fmt->repeat;
+
+		if (index < 0)
+			break;
+		if (!TEST_BIT(mask, number))
+			continue;
+
+		if (prev && index == iio_channel_get_index(prev)) {
+			prev = chn;
+			continue;
+		}
+
+		if (size % len)
+			size += 2 * len - (size % len);
+		else
+			size += len;
+
+		prev = chn;
+	}
+
+	return size;
+}
+
 static int open_dev_helper(struct parser_pdata *pdata, struct iio_device *dev,
 		size_t samples_count, const char *mask, bool cyclic)
 {
@@ -880,7 +972,7 @@ static int open_dev_helper(struct parser_pdata *pdata, struct iio_device *dev,
 	if (!dev)
 		return -ENODEV;
 
-	nb_channels = dev->nb_channels;
+	nb_channels = iio_device_get_channels_count(dev);
 	if (len != ((nb_channels + 31) / 32) * 8)
 		return -EINVAL;
 
@@ -896,7 +988,7 @@ static int open_dev_helper(struct parser_pdata *pdata, struct iio_device *dev,
 	thd->mask = words;
 	thd->nb = 0;
 	thd->samples_count = samples_count;
-	thd->sample_size = iio_device_get_sample_size_mask(dev, words, len);
+	thd->sample_size = get_dev_sample_size_mask(dev, words, len);
 	thd->pdata = pdata;
 	thd->dev = dev;
 	thd->eventfd = eventfd(0, EFD_CLOEXEC | EFD_NONBLOCK);
@@ -938,9 +1030,12 @@ retry:
 			 * This is not pretty but it works.
 			 */
 			if (cyclic_retry) {
-			       cyclic_retry--;
-			       usleep(100);
-			       goto retry;
+				struct timespec wait;
+				wait.tv_sec = 0;
+				wait.tv_nsec = (100 * 1000);
+				cyclic_retry--;
+				nanosleep(&wait, &wait);
+				goto retry;
 			}
 
 			ret = -EBUSY;
@@ -956,7 +1051,7 @@ retry:
 			SLIST_INSERT_HEAD(&entry->thdlist_head, thd, dev_list_entry);
 			thd->entry = entry;
 			entry->update_mask = true;
-			DEBUG("Added thread to client list\n");
+			IIO_DEBUG("Added thread to client list\n");
 
 			pthread_cond_signal(&entry->rw_ready_cond);
 
@@ -1002,7 +1097,7 @@ retry:
 	SLIST_INIT(&entry->thdlist_head);
 	SLIST_INSERT_HEAD(&entry->thdlist_head, thd, dev_list_entry);
 	thd->entry = entry;
-	DEBUG("Added thread to client list\n");
+	IIO_DEBUG("Added thread to client list\n");
 
 	pthread_mutex_init(&entry->thdlist_lock, NULL);
 	pthread_cond_init(&entry->rw_ready_cond, NULL);
@@ -1013,7 +1108,7 @@ retry:
 		goto err_free_entry_mask;
 	}
 
-	DEBUG("Adding new device thread to device list\n");
+	IIO_DEBUG("Adding new device thread to device list\n");
 	iio_device_set_data(dev, entry);
 	pthread_mutex_unlock(&devlist_lock);
 
@@ -1088,24 +1183,35 @@ ssize_t rw_dev(struct parser_pdata *pdata, struct iio_device *dev,
 }
 
 ssize_t read_dev_attr(struct parser_pdata *pdata, struct iio_device *dev,
-		const char *attr, bool is_debug)
+		const char *attr, enum iio_attr_type type)
 {
 	/* We use a very large buffer here, as if attr is NULL all the
 	 * attributes will be read, which may represents a few kilobytes worth
 	 * of data. */
 	char buf[0x10000];
-	ssize_t ret;
+	ssize_t ret = -EINVAL;
 
 	if (!dev) {
 		print_value(pdata, -ENODEV);
 		return -ENODEV;
 	}
 
-	if (is_debug)
-		ret = iio_device_debug_attr_read(dev,
+	switch (type) {
+		case IIO_ATTR_TYPE_DEVICE:
+			ret = iio_device_attr_read(dev, attr, buf, sizeof(buf) - 1);
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			ret = iio_device_debug_attr_read(dev,
 				attr, buf, sizeof(buf) - 1);
-	else
-		ret = iio_device_attr_read(dev, attr, buf, sizeof(buf) - 1);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			ret = iio_device_buffer_attr_read(dev,
+							attr, buf, sizeof(buf) - 1);
+			break;
+		default:
+			ret = -EINVAL;
+			break;
+	}
 	print_value(pdata, ret);
 	if (ret < 0)
 		return ret;
@@ -1115,7 +1221,7 @@ ssize_t read_dev_attr(struct parser_pdata *pdata, struct iio_device *dev,
 }
 
 ssize_t write_dev_attr(struct parser_pdata *pdata, struct iio_device *dev,
-		const char *attr, size_t len, bool is_debug)
+		const char *attr, size_t len, enum iio_attr_type type)
 {
 	ssize_t ret = -ENOMEM;
 	char *buf;
@@ -1133,10 +1239,20 @@ ssize_t write_dev_attr(struct parser_pdata *pdata, struct iio_device *dev,
 	if (ret < 0)
 		goto err_free_buffer;
 
-	if (is_debug)
-		ret = iio_device_debug_attr_write_raw(dev, attr, buf, len);
-	else
-		ret = iio_device_attr_write_raw(dev, attr, buf, len);
+	switch (type) {
+		case IIO_ATTR_TYPE_DEVICE:
+			ret = iio_device_attr_write_raw(dev, attr, buf, len);
+			break;
+		case IIO_ATTR_TYPE_DEBUG:
+			ret = iio_device_debug_attr_write_raw(dev, attr, buf, len);
+			break;
+		case IIO_ATTR_TYPE_BUFFER:
+			ret = iio_device_buffer_attr_write_raw(dev, attr, buf, len);
+			break;
+		default:
+			ret = -EINVAL;
+			break;
+	}
 
 err_free_buffer:
 	free(buf);
@@ -1223,12 +1339,13 @@ ssize_t get_trigger(struct parser_pdata *pdata, struct iio_device *dev)
 
 	ret = iio_device_get_trigger(dev, &trigger);
 	if (!ret && trigger) {
+		const char *name = iio_device_get_name(trigger);
 		char buf[256];
 
-		ret = strlen(trigger->name);
+		ret = strlen(name);
 		print_value(pdata, ret);
 
-		snprintf(buf, sizeof(buf), "%s\n", trigger->name);
+		snprintf(buf, sizeof(buf), "%s\n", name);
 		ret = write_all(pdata, buf, ret + 1);
 	} else {
 		print_value(pdata, ret);
@@ -1246,24 +1363,49 @@ int set_timeout(struct parser_pdata *pdata, unsigned int timeout)
 int set_buffers_count(struct parser_pdata *pdata,
 		struct iio_device *dev, long value)
 {
+	unsigned int i, nb = (unsigned int) value;
+	struct timespec wait;
 	int ret = -EINVAL;
 
-	if (value >= 1)
-		ret = iio_device_set_kernel_buffers_count(
-				dev, (unsigned int) value);
+	if (!dev) {
+		ret = -ENODEV;
+		goto err_print_value;
+	}
 
+	if (nb >= 1) {
+		/*
+		 * Avoid the same race condition described in open_dev_helper().
+		 * We must be sure that the buffer has not been enabled in order
+		 * to set the number of kernel buffers.
+		 */
+		for (i = 0; i < 500; i++) {
+			ret = iio_device_set_kernel_buffers_count(dev, nb);
+			if (ret != -EBUSY)
+				break;
+
+			wait.tv_sec = 0;
+			wait.tv_nsec = (100 * 1000);
+			do {
+				ret = nanosleep(&wait, &wait);
+			} while (ret == -1 && errno == EINTR);
+		}
+	}
+err_print_value:
 	print_value(pdata, ret);
 	return ret;
 }
 
 ssize_t read_line(struct parser_pdata *pdata, char *buf, size_t len)
 {
+	size_t bytes_read = 0;
 	ssize_t ret;
+	bool found;
+
+	if (pdata->is_usb)
+	      return pdata->readfd(pdata, buf, len);
 
 	if (pdata->fd_in_is_socket) {
 		struct pollfd pfd[2];
-		bool found;
-		size_t bytes_read = 0;
 
 		pfd[0].fd = pdata->fd_in;
 		pfd[0].events = POLLIN | POLLRDHUP;
@@ -1306,21 +1448,31 @@ ssize_t read_line(struct parser_pdata *pdata, char *buf, size_t len)
 
 			bytes_read += to_trunc;
 		} while (!found && len);
-
-		/* No \n found? Just garbage data */
-		if (!found)
-			ret = -EIO;
-		else
-			ret = bytes_read;
 	} else {
-		ret = pdata->readfd(pdata, buf, len);
+		while (len) {
+			ret = pdata->readfd(pdata, buf, 1);
+			if (ret < 0)
+			      return ret;
+
+			bytes_read++;
+
+			if (*buf == '\n')
+			      break;
+
+			len--;
+			buf++;
+		}
+
+		found = !!len;
 	}
 
-	return ret;
+	return found ? (ssize_t) bytes_read : -EIO;
 }
 
 void interpreter(struct iio_context *ctx, int fd_in, int fd_out, bool verbose,
-	bool is_socket, bool use_aio, struct thread_pool *pool)
+		 bool is_socket, bool is_usb, bool use_aio,
+		 struct thread_pool *pool, const void *xml_zstd,
+		 size_t xml_zstd_len)
 {
 	yyscan_t scanner;
 	struct parser_pdata pdata;
@@ -1334,8 +1486,12 @@ void interpreter(struct iio_context *ctx, int fd_in, int fd_out, bool verbose,
 	pdata.verbose = verbose;
 	pdata.pool = pool;
 
+	pdata.xml_zstd = xml_zstd;
+	pdata.xml_zstd_len = xml_zstd_len;
+
 	pdata.fd_in_is_socket = is_socket;
 	pdata.fd_out_is_socket = is_socket;
+	pdata.is_usb = is_usb;
 
 	SLIST_INIT(&pdata.thdlist_head);
 
@@ -1348,7 +1504,7 @@ void interpreter(struct iio_context *ctx, int fd_in, int fd_out, bool verbose,
 		pdata.aio_eventfd = eventfd(0, EFD_CLOEXEC | EFD_NONBLOCK);
 		if (pdata.aio_eventfd < 0) {
 			iio_strerror(errno, err_str, sizeof(err_str));
-			ERROR("Failed to create AIO eventfd: %s\n", err_str);
+			IIO_ERROR("Failed to create AIO eventfd: %s\n", err_str);
 			return;
 		}
 
@@ -1356,7 +1512,7 @@ void interpreter(struct iio_context *ctx, int fd_in, int fd_out, bool verbose,
 		ret = io_setup(1, &pdata.aio_ctx);
 		if (ret < 0) {
 			iio_strerror(-ret, err_str, sizeof(err_str));
-			ERROR("Failed to create AIO context: %s\n", err_str);
+			IIO_ERROR("Failed to create AIO context: %s\n", err_str);
 			close(pdata.aio_eventfd);
 			return;
 		}
@@ -1380,8 +1536,8 @@ void interpreter(struct iio_context *ctx, int fd_in, int fd_out, bool verbose,
 	yylex_destroy(scanner);
 
 	/* Close all opened devices */
-	for (i = 0; i < ctx->nb_devices; i++)
-		close_dev_helper(&pdata, ctx->devices[i]);
+	for (i = 0; i < iio_context_get_devices_count(ctx); i++)
+		close_dev_helper(&pdata, iio_context_get_device(ctx, i));
 
 #if WITH_AIO
 	if (use_aio) {
