@@ -1,20 +1,10 @@
+// SPDX-License-Identifier: LGPL-2.1-or-later
 /*
  * libiio - Library for interfacing industrial I/O (IIO) devices
  *
- * Copyright (C) 2015 Analog Devices, Inc.
+ * Copyright (C) 2015 - 2020 Analog Devices, Inc.
  * Author: Paul Cercueil <paul.cercueil@analog.com>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * */
+ */
 
 #include "iio-lock.h"
 #include "iio-private.h"
@@ -22,14 +12,10 @@
 
 #include <ctype.h>
 #include <errno.h>
-#include <libusb-1.0/libusb.h>
+#include <libusb.h>
 #include <stdbool.h>
 #include <string.h>
 
-#ifdef ERROR
-#undef ERROR
-#endif
-
 #include "debug.h"
 
 #define DEFAULT_TIMEOUT_MS 5000
@@ -41,13 +27,13 @@
 
 struct iio_usb_ep_couple {
 	unsigned char addr_in, addr_out;
-	unsigned int pipe_id;
+	uint16_t pipe_id;
 	bool in_use;
 
 	struct iio_mutex *lock;
 };
 
-struct iio_usb_io_context {
+struct iiod_client_pdata {
 	struct iio_usb_ep_couple *ep;
 
 	struct iio_mutex *lock;
@@ -58,29 +44,26 @@ struct iio_usb_io_context {
 struct iio_context_pdata {
 	libusb_context *ctx;
 	libusb_device_handle *hdl;
-	unsigned int interface;
+	uint16_t intrfc;
 
 	struct iiod_client *iiod_client;
 
-	/* Lock for non-streaming operations */
-	struct iio_mutex *lock;
-
 	/* Lock for endpoint reservation */
 	struct iio_mutex *ep_lock;
 
 	struct iio_usb_ep_couple *io_endpoints;
-	unsigned int nb_ep_couples;
+	uint16_t nb_ep_couples;
 
 	unsigned int timeout_ms;
 
-	struct iio_usb_io_context io_ctx;
+	struct iiod_client_pdata io_ctx;
 };
 
 struct iio_device_pdata {
 	struct iio_mutex *lock;
 
 	bool opened;
-	struct iio_usb_io_context io_ctx;
+	struct iiod_client_pdata io_ctx;
 };
 
 static const unsigned int libusb_to_errno_codes[] = {
@@ -119,7 +102,7 @@ static unsigned int libusb_to_errno(int error)
 	}
 }
 
-static int usb_io_context_init(struct iio_usb_io_context *io_ctx)
+static int usb_io_context_init(struct iiod_client_pdata *io_ctx)
 {
 	io_ctx->lock = iio_mutex_create();
 	if (!io_ctx->lock)
@@ -128,7 +111,7 @@ static int usb_io_context_init(struct iio_usb_io_context *io_ctx)
 	return 0;
 }
 
-static void usb_io_context_exit(struct iio_usb_io_context *io_ctx)
+static void usb_io_context_exit(struct iiod_client_pdata *io_ctx)
 {
 	if (io_ctx->lock) {
 		iio_mutex_destroy(io_ctx->lock);
@@ -139,8 +122,10 @@ static void usb_io_context_exit(struct iio_usb_io_context *io_ctx)
 static int usb_get_version(const struct iio_context *ctx,
 		unsigned int *major, unsigned int *minor, char git_tag[8])
 {
-	return iiod_client_get_version(ctx->pdata->iiod_client,
-			&ctx->pdata->io_ctx, major, minor, git_tag);
+	struct iio_context_pdata *pdata = iio_context_get_pdata(ctx);
+
+	return iiod_client_get_version(pdata->iiod_client,
+			&pdata->io_ctx, major, minor, git_tag);
 }
 
 static unsigned int usb_calculate_remote_timeout(unsigned int timeout)
@@ -150,7 +135,7 @@ static unsigned int usb_calculate_remote_timeout(unsigned int timeout)
 	return timeout / 2;
 }
 
-#define USB_PIPE_CTRL_TIMEOUT 200 /* These should not take long */
+#define USB_PIPE_CTRL_TIMEOUT 1000 /* These should not take long */
 
 #define IIO_USD_CMD_RESET_PIPES 0
 #define IIO_USD_CMD_OPEN_PIPE 1
@@ -159,34 +144,63 @@ static unsigned int usb_calculate_remote_timeout(unsigned int timeout)
 static int usb_reset_pipes(struct iio_context_pdata *pdata)
 {
 	int ret;
-
-	ret = libusb_control_transfer(pdata->hdl, LIBUSB_REQUEST_TYPE_VENDOR |
-		LIBUSB_RECIPIENT_INTERFACE, IIO_USD_CMD_RESET_PIPES,
-		0, pdata->interface, NULL, 0, USB_PIPE_CTRL_TIMEOUT);
+/*
+	int libusb_control_transfer(libusb_device_handle *devh,
+			uint8_t bmRequestType,
+			uint8_t bRequest,
+			uint16_t wValue,
+			uint16_t wIndex,
+			unsigned char *data,
+			uint16_t wLength,
+			unsigned int timeout)
+*/
+	ret = libusb_control_transfer(pdata->hdl,
+			LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_INTERFACE,
+			IIO_USD_CMD_RESET_PIPES,
+			0,
+			pdata->intrfc,
+			NULL,
+			0,
+			USB_PIPE_CTRL_TIMEOUT);
 	if (ret < 0)
 		return -(int) libusb_to_errno(ret);
 	return 0;
 }
 
-static int usb_open_pipe(struct iio_context_pdata *pdata, unsigned int pipe_id)
+static int usb_open_pipe(struct iio_context_pdata *pdata, uint16_t pipe_id)
 {
 	int ret;
-
-	ret = libusb_control_transfer(pdata->hdl, LIBUSB_REQUEST_TYPE_VENDOR |
-		LIBUSB_RECIPIENT_INTERFACE, IIO_USD_CMD_OPEN_PIPE,
-		pipe_id, pdata->interface, NULL, 0, USB_PIPE_CTRL_TIMEOUT);
+/*
+	libusb_device_handle *devh,
+	uint8_t bmRequestType,
+	uint8_t bRequest,
+	uint16_t wValue,
+	uint16_t wIndex,
+	unsigned char *data,
+	uint16_t wLength,
+	unsigned int timeout)
+*/
+	ret = libusb_control_transfer(
+			pdata->hdl,
+			LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_INTERFACE,
+			IIO_USD_CMD_OPEN_PIPE,
+			pipe_id,
+			pdata->intrfc,
+			NULL,
+			0,
+			USB_PIPE_CTRL_TIMEOUT);
 	if (ret < 0)
 		return -(int) libusb_to_errno(ret);
 	return 0;
 }
 
-static int usb_close_pipe(struct iio_context_pdata *pdata, unsigned int pipe_id)
+static int usb_close_pipe(struct iio_context_pdata *pdata, uint16_t pipe_id)
 {
 	int ret;
 
 	ret = libusb_control_transfer(pdata->hdl, LIBUSB_REQUEST_TYPE_VENDOR |
 		LIBUSB_RECIPIENT_INTERFACE, IIO_USD_CMD_CLOSE_PIPE,
-		pipe_id, pdata->interface, NULL, 0, USB_PIPE_CTRL_TIMEOUT);
+		pipe_id, pdata->intrfc, NULL, 0, USB_PIPE_CTRL_TIMEOUT);
 	if (ret < 0)
 		return -(int) libusb_to_errno(ret);
 	return 0;
@@ -194,7 +208,7 @@ static int usb_close_pipe(struct iio_context_pdata *pdata, unsigned int pipe_id)
 
 static int usb_reserve_ep_unlocked(const struct iio_device *dev)
 {
-	struct iio_context_pdata *pdata = dev->ctx->pdata;
+	struct iio_context_pdata *pdata = iio_context_get_pdata(dev->ctx);
 	unsigned int i;
 
 	for (i = 0; i < pdata->nb_ep_couples; i++) {
@@ -214,7 +228,7 @@ static int usb_reserve_ep_unlocked(const struct iio_device *dev)
 
 static void usb_free_ep_unlocked(const struct iio_device *dev)
 {
-	struct iio_context_pdata *pdata = dev->ctx->pdata;
+	struct iio_context_pdata *pdata = iio_context_get_pdata(dev->ctx);
 	unsigned int i;
 
 	for (i = 0; i < pdata->nb_ep_couples; i++) {
@@ -230,7 +244,7 @@ static void usb_free_ep_unlocked(const struct iio_device *dev)
 static int usb_open(const struct iio_device *dev,
 		size_t samples_count, bool cyclic)
 {
-	struct iio_context_pdata *ctx_pdata = dev->ctx->pdata;
+	struct iio_context_pdata *ctx_pdata = iio_context_get_pdata(dev->ctx);
 	struct iio_device_pdata *pdata = dev->pdata;
 	int ret = -EBUSY;
 
@@ -250,7 +264,7 @@ static int usb_open(const struct iio_device *dev,
 		char err_str[1024];
 
 		iio_strerror(-ret, err_str, sizeof(err_str));
-		ERROR("Failed to open pipe: %s\n", err_str);
+		IIO_ERROR("Failed to open pipe: %s\n", err_str);
 		usb_free_ep_unlocked(dev);
 		goto out_unlock;
 	}
@@ -284,7 +298,7 @@ out_unlock:
 
 static int usb_close(const struct iio_device *dev)
 {
-	struct iio_context_pdata *ctx_pdata = dev->ctx->pdata;
+	struct iio_context_pdata *ctx_pdata = iio_context_get_pdata(dev->ctx);
 	struct iio_device_pdata *pdata = dev->pdata;
 	int ret = -EBADF;
 
@@ -311,11 +325,12 @@ out_unlock:
 static ssize_t usb_read(const struct iio_device *dev, void *dst, size_t len,
 		uint32_t *mask, size_t words)
 {
+	struct iio_context_pdata *ctx_pdata = iio_context_get_pdata(dev->ctx);
 	struct iio_device_pdata *pdata = dev->pdata;
 	ssize_t ret;
 
 	iio_mutex_lock(pdata->lock);
-	ret = iiod_client_read_unlocked(dev->ctx->pdata->iiod_client,
+	ret = iiod_client_read_unlocked(ctx_pdata->iiod_client,
 			&pdata->io_ctx, dev, dst, len, mask, words);
 	iio_mutex_unlock(pdata->lock);
 
@@ -325,11 +340,12 @@ static ssize_t usb_read(const struct iio_device *dev, void *dst, size_t len,
 static ssize_t usb_write(const struct iio_device *dev,
 		const void *src, size_t len)
 {
+	struct iio_context_pdata *ctx_pdata = iio_context_get_pdata(dev->ctx);
 	struct iio_device_pdata *pdata = dev->pdata;
 	ssize_t ret;
 
 	iio_mutex_lock(pdata->lock);
-	ret = iiod_client_write_unlocked(dev->ctx->pdata->iiod_client,
+	ret = iiod_client_write_unlocked(ctx_pdata->iiod_client,
 			&pdata->io_ctx, dev, src, len);
 	iio_mutex_unlock(pdata->lock);
 
@@ -337,29 +353,29 @@ static ssize_t usb_write(const struct iio_device *dev,
 }
 
 static ssize_t usb_read_dev_attr(const struct iio_device *dev,
-		const char *attr, char *dst, size_t len, bool is_debug)
+		const char *attr, char *dst, size_t len, enum iio_attr_type type)
 {
-	struct iio_context_pdata *pdata = dev->ctx->pdata;
+	struct iio_context_pdata *pdata = iio_context_get_pdata(dev->ctx);
 
 	return iiod_client_read_attr(pdata->iiod_client,
 			&pdata->io_ctx, dev, NULL, attr,
-			dst, len, is_debug);
+			dst, len, type);
 }
 
 static ssize_t usb_write_dev_attr(const struct iio_device *dev,
-		const char *attr, const char *src, size_t len, bool is_debug)
+		const char *attr, const char *src, size_t len, enum iio_attr_type type)
 {
-	struct iio_context_pdata *pdata = dev->ctx->pdata;
+	struct iio_context_pdata *pdata = iio_context_get_pdata(dev->ctx);
 
 	return iiod_client_write_attr(pdata->iiod_client,
 			&pdata->io_ctx, dev, NULL, attr,
-			src, len, is_debug);
+			src, len, type);
 }
 
 static ssize_t usb_read_chn_attr(const struct iio_channel *chn,
 		const char *attr, char *dst, size_t len)
 {
-	struct iio_context_pdata *pdata = chn->dev->ctx->pdata;
+	struct iio_context_pdata *pdata = iio_context_get_pdata(chn->dev->ctx);
 
 	return iiod_client_read_attr(pdata->iiod_client,
 			&pdata->io_ctx, chn->dev, chn, attr,
@@ -369,7 +385,7 @@ static ssize_t usb_read_chn_attr(const struct iio_channel *chn,
 static ssize_t usb_write_chn_attr(const struct iio_channel *chn,
 		const char *attr, const char *src, size_t len)
 {
-	struct iio_context_pdata *pdata = chn->dev->ctx->pdata;
+	struct iio_context_pdata *pdata = iio_context_get_pdata(chn->dev->ctx);
 
 	return iiod_client_write_attr(pdata->iiod_client,
 			&pdata->io_ctx, chn->dev, chn, attr,
@@ -379,7 +395,7 @@ static ssize_t usb_write_chn_attr(const struct iio_channel *chn,
 static int usb_set_kernel_buffers_count(const struct iio_device *dev,
 		unsigned int nb_blocks)
 {
-	struct iio_context_pdata *pdata = dev->ctx->pdata;
+	struct iio_context_pdata *pdata = iio_context_get_pdata(dev->ctx);
 
 	return iiod_client_set_kernel_buffers_count(pdata->iiod_client,
 			&pdata->io_ctx, dev, nb_blocks);
@@ -387,7 +403,7 @@ static int usb_set_kernel_buffers_count(const struct iio_device *dev,
 
 static int usb_set_timeout(struct iio_context *ctx, unsigned int timeout)
 {
-	struct iio_context_pdata *pdata = ctx->pdata;
+	struct iio_context_pdata *pdata = iio_context_get_pdata(ctx);
 	unsigned int remote_timeout = usb_calculate_remote_timeout(timeout);
 	int ret;
 
@@ -399,50 +415,68 @@ static int usb_set_timeout(struct iio_context *ctx, unsigned int timeout)
 	return ret;
 }
 
+static int usb_get_trigger(const struct iio_device *dev,
+                const struct iio_device **trigger)
+{
+	struct iio_context_pdata *pdata = iio_context_get_pdata(dev->ctx);
+
+	return iiod_client_get_trigger(pdata->iiod_client,
+			&pdata->io_ctx, dev, trigger);
+}
+
+static int usb_set_trigger(const struct iio_device *dev,
+                const struct iio_device *trigger)
+{
+	struct iio_context_pdata *pdata = iio_context_get_pdata(dev->ctx);
+
+	return iiod_client_set_trigger(pdata->iiod_client,
+			&pdata->io_ctx, dev, trigger);
+}
+
+
 static void usb_shutdown(struct iio_context *ctx)
 {
+	struct iio_context_pdata *pdata = iio_context_get_pdata(ctx);
+	unsigned int nb_devices = iio_context_get_devices_count(ctx);
 	unsigned int i;
 
-	usb_io_context_exit(&ctx->pdata->io_ctx);
+	usb_io_context_exit(&pdata->io_ctx);
 
-	for (i = 0; i < ctx->nb_devices; i++)
-		usb_close(ctx->devices[i]);
+	for (i = 0; i < nb_devices; i++)
+		usb_close(iio_context_get_device(ctx, i));
 
-	iio_mutex_destroy(ctx->pdata->lock);
-	iio_mutex_destroy(ctx->pdata->ep_lock);
+	iio_mutex_destroy(pdata->ep_lock);
 
-	for (i = 0; i < ctx->pdata->nb_ep_couples; i++)
-		if (ctx->pdata->io_endpoints[i].lock)
-			iio_mutex_destroy(ctx->pdata->io_endpoints[i].lock);
-	if (ctx->pdata->io_endpoints)
-		free(ctx->pdata->io_endpoints);
+	for (i = 0; i < pdata->nb_ep_couples; i++)
+		if (pdata->io_endpoints[i].lock)
+			iio_mutex_destroy(pdata->io_endpoints[i].lock);
+	free(pdata->io_endpoints);
 
-	for (i = 0; i < ctx->nb_devices; i++) {
-		struct iio_device *dev = ctx->devices[i];
+	for (i = 0; i < nb_devices; i++) {
+		struct iio_device *dev = iio_context_get_device(ctx, i);
 
 		usb_io_context_exit(&dev->pdata->io_ctx);
 		free(dev->pdata);
 	}
 
-	iiod_client_destroy(ctx->pdata->iiod_client);
+	iiod_client_destroy(pdata->iiod_client);
 
-	usb_reset_pipes(ctx->pdata); /* Close everything */
+	usb_reset_pipes(pdata); /* Close everything */
 
-	libusb_close(ctx->pdata->hdl);
-	libusb_exit(ctx->pdata->ctx);
-	free(ctx->pdata);
+	libusb_close(pdata->hdl);
+	libusb_exit(pdata->ctx);
 }
 
 static int iio_usb_match_interface(const struct libusb_config_descriptor *desc,
-		struct libusb_device_handle *hdl, unsigned int interface)
+		struct libusb_device_handle *hdl, unsigned int intrfc)
 {
 	const struct libusb_interface *iface;
 	unsigned int i;
 
-	if (interface >= desc->bNumInterfaces)
+	if (intrfc >= desc->bNumInterfaces)
 		return -EINVAL;
 
-	iface = &desc->interface[interface];
+	iface = &desc->interface[intrfc];
 
 	for (i = 0; i < (unsigned int) iface->num_altsetting; i++) {
 		const struct libusb_interface_descriptor *idesc =
@@ -467,7 +501,7 @@ static int iio_usb_match_interface(const struct libusb_config_descriptor *desc,
 
 static int iio_usb_match_device(struct libusb_device *dev,
 		struct libusb_device_handle *hdl,
-		unsigned int *interface)
+		unsigned int *intrfc)
 {
 	struct libusb_config_descriptor *desc;
 	unsigned int i;
@@ -485,11 +519,11 @@ static int iio_usb_match_device(struct libusb_device *dev,
 	if (ret < 0)
 		return ret;
 
-	DEBUG("Found IIO interface on device %u:%u using interface %u\n",
+	IIO_DEBUG("Found IIO interface on device %u:%u using interface %u\n",
 			libusb_get_bus_number(dev),
 			libusb_get_device_address(dev), i - 1);
 
-	*interface = i - 1;
+	*intrfc = i - 1;
 	return ret;
 }
 
@@ -514,6 +548,8 @@ static const struct iio_backend_ops usb_ops = {
 	.read_channel_attr = usb_read_chn_attr,
 	.write_device_attr = usb_write_dev_attr,
 	.write_channel_attr = usb_write_chn_attr,
+	.get_trigger = usb_get_trigger,
+	.set_trigger = usb_set_trigger,
 	.set_kernel_buffers_count = usb_set_kernel_buffers_count,
 	.set_timeout = usb_set_timeout,
 	.shutdown = usb_shutdown,
@@ -528,11 +564,11 @@ static void LIBUSB_CALL sync_transfer_cb(struct libusb_transfer *transfer)
 }
 
 static int usb_sync_transfer(struct iio_context_pdata *pdata,
-	struct iio_usb_io_context *io_ctx, unsigned int ep_type,
+	struct iiod_client_pdata *io_ctx, unsigned int ep_type,
 	char *data, size_t len, int *transferred)
 {
-	unsigned int ep;
-	struct libusb_transfer *transfer;
+	unsigned char ep;
+	struct libusb_transfer *transfer = NULL;
 	int completed = 0;
 	int ret;
 
@@ -633,7 +669,8 @@ unlock:
 }
 
 static ssize_t write_data_sync(struct iio_context_pdata *pdata,
-		void *ep, const char *data, size_t len)
+			       struct iiod_client_pdata *ep,
+			       const char *data, size_t len)
 {
 	int transferred, ret;
 
@@ -646,7 +683,8 @@ static ssize_t write_data_sync(struct iio_context_pdata *pdata,
 }
 
 static ssize_t read_data_sync(struct iio_context_pdata *pdata,
-		void *ep, char *buf, size_t len)
+			      struct iiod_client_pdata *ep,
+			      char *buf, size_t len)
 {
 	int transferred, ret;
 
@@ -690,10 +728,12 @@ static int usb_verify_eps(const struct libusb_interface_descriptor *iface)
 static int usb_populate_context_attrs(struct iio_context *ctx,
 		libusb_device *dev, libusb_device_handle *hdl)
 {
+	struct iio_context_pdata *pdata = iio_context_get_pdata(ctx);
 	struct libusb_device_descriptor dev_desc;
 	char buffer[64];
 	unsigned int i;
 	int ret;
+	char uri[sizeof("usb:127.255.255")];
 
 	struct {
 		const char *attr;
@@ -709,6 +749,13 @@ static int usb_populate_context_attrs(struct iio_context *ctx,
 	attrs[2].attr = "usb,serial";
 	attrs[2].idx = dev_desc.iSerialNumber;
 
+	iio_snprintf(uri, sizeof(uri), "usb:%d.%d.%u",
+		libusb_get_bus_number(dev), libusb_get_device_address(dev),
+		(uint8_t)pdata->intrfc);
+	ret = iio_context_add_attr(ctx, "uri", uri);
+	if (ret < 0)
+		return ret;
+
 	iio_snprintf(buffer, sizeof(buffer), "%04hx", dev_desc.idVendor);
 	ret = iio_context_add_attr(ctx, "usb,idVendor", buffer);
 	if (ret < 0)
@@ -740,14 +787,30 @@ static int usb_populate_context_attrs(struct iio_context *ctx,
 		}
 	}
 
+#ifdef HAS_LIBUSB_GETVERSION
+	/*
+	 * libusb_get_version was added 2012-04-17: v1.0.10,
+	 * before LIBUSB_API_VERSION was added - Jan 8, 2014
+	 * so, you can't use that to determine if it is here
+	 */
+	{
+	struct libusb_version const *ver = libusb_get_version();
+	iio_snprintf(buffer, sizeof(buffer), "%i.%i.%i.%i%s",
+			ver->major, ver->minor, ver->micro,
+			ver->nano, ver->rc);
+	ret = iio_context_add_attr(ctx, "usb,libusb", buffer);
+	if (ret < 0)
+		return ret;
+	}
+#endif
 	return 0;
 }
 
-struct iio_context * usb_create_context(unsigned int bus,
-		unsigned int address, unsigned int interface)
+static struct iio_context * usb_create_context(unsigned int bus,
+					       uint16_t address, uint16_t intrfc)
 {
 	libusb_context *usb_ctx;
-	libusb_device_handle *hdl;
+	libusb_device_handle *hdl = NULL;
 	const struct libusb_interface_descriptor *iface;
 	libusb_device *usb_dev;
 	struct libusb_config_descriptor *conf_desc;
@@ -755,34 +818,26 @@ struct iio_context * usb_create_context(unsigned int bus,
 	struct iio_context *ctx;
 	struct iio_context_pdata *pdata;
 	char err_str[1024];
-	unsigned int i;
+	uint16_t i;
 	int ret;
 
 	pdata = zalloc(sizeof(*pdata));
 	if (!pdata) {
-		ERROR("Unable to allocate pdata\n");
+		IIO_ERROR("Unable to allocate pdata\n");
 		ret = -ENOMEM;
 		goto err_set_errno;
 	}
 
-	pdata->lock = iio_mutex_create();
-	if (!pdata->lock) {
-		ERROR("Unable to create mutex\n");
+	pdata->ep_lock = iio_mutex_create();
+	if (!pdata->ep_lock) {
+		IIO_ERROR("Unable to create mutex\n");
 		ret = -ENOMEM;
 		goto err_free_pdata;
 	}
 
-	pdata->ep_lock = iio_mutex_create();
-	if (!pdata->ep_lock) {
-		ERROR("Unable to create mutex\n");
-		ret = -ENOMEM;
-		goto err_destroy_mutex;
-	}
-
-	pdata->iiod_client = iiod_client_new(pdata, pdata->lock,
-			&usb_iiod_client_ops);
+	pdata->iiod_client = iiod_client_new(pdata, &usb_iiod_client_ops);
 	if (!pdata->iiod_client) {
-		ERROR("Unable to create IIOD client\n");
+		IIO_ERROR("Unable to create IIOD client\n");
 		ret = -errno;
 		goto err_destroy_ep_mutex;
 	}
@@ -790,11 +845,16 @@ struct iio_context * usb_create_context(unsigned int bus,
 	ret = libusb_init(&usb_ctx);
 	if (ret) {
 		ret = -(int) libusb_to_errno(ret);
-		ERROR("Unable to init libusb: %i\n", ret);
+		IIO_ERROR("Unable to init libusb: %i\n", ret);
 		goto err_destroy_iiod_client;
 	}
 
-	libusb_get_device_list(usb_ctx, &device_list);
+	ret = (int) libusb_get_device_list(usb_ctx, &device_list);
+	if (ret < 0) {
+		ret = -(int) libusb_to_errno(ret);
+		IIO_ERROR("Unable to get usb device list: %i\n", ret);
+		goto err_libusb_exit;
+	}
 
 	usb_dev = NULL;
 
@@ -815,7 +875,7 @@ struct iio_context * usb_create_context(unsigned int bus,
 			 * we find such a device skip it and keep looking.
 			 */
 			if (ret == LIBUSB_ERROR_NOT_SUPPORTED) {
-				WARNING("Skipping broken USB device. Please upgrade libusb.\n");
+				IIO_WARNING("Skipping broken USB device. Please upgrade libusb.\n");
 				usb_dev = NULL;
 				continue;
 			}
@@ -826,14 +886,14 @@ struct iio_context * usb_create_context(unsigned int bus,
 
 	libusb_free_device_list(device_list, true);
 
-	if (!usb_dev) {
+	if (!usb_dev || !hdl) {
 		ret = -ENODEV;
 		goto err_libusb_exit;
 	}
 
 	if (ret) {
 		ret = -(int) libusb_to_errno(ret);
-		ERROR("Unable to open device\n");
+		IIO_ERROR("Unable to open device\n");
 		goto err_libusb_exit;
 	}
 
@@ -841,36 +901,39 @@ struct iio_context * usb_create_context(unsigned int bus,
 	libusb_set_auto_detach_kernel_driver(hdl, true);
 #endif
 
-	ret = libusb_claim_interface(hdl, interface);
+	ret = libusb_claim_interface(hdl, intrfc);
 	if (ret) {
 		ret = -(int) libusb_to_errno(ret);
-		ERROR("Unable to claim interface %u: %i\n", interface, ret);
+		iio_strerror(-ret, err_str, sizeof(err_str));
+		IIO_ERROR("Unable to claim interface %u:%u:%u: %s\n",
+		      bus, address, intrfc, err_str);
 		goto err_libusb_close;
 	}
 
 	ret = libusb_get_active_config_descriptor(usb_dev, &conf_desc);
 	if (ret) {
 		ret = -(int) libusb_to_errno(ret);
-		ERROR("Unable to get config descriptor: %i\n", ret);
+		iio_strerror(-ret, err_str, sizeof(err_str));
+		IIO_ERROR("Unable to get config descriptor: %s\n", err_str);
 		goto err_libusb_close;
 	}
 
-	iface = &conf_desc->interface[interface].altsetting[0];
+	iface = &conf_desc->interface[intrfc].altsetting[0];
 
 	ret = usb_verify_eps(iface);
 	if (ret) {
-		ERROR("Invalid configuration of endpoints\n");
+		IIO_ERROR("Invalid configuration of endpoints\n");
 		goto err_free_config_descriptor;
 	}
 
 	pdata->nb_ep_couples = iface->bNumEndpoints / 2;
 
-	DEBUG("Found %hhu usable i/o endpoint couples\n", pdata->nb_ep_couples);
+	IIO_DEBUG("Found %hu usable i/o endpoint couples\n", pdata->nb_ep_couples);
 
 	pdata->io_endpoints = calloc(pdata->nb_ep_couples,
 			sizeof(*pdata->io_endpoints));
 	if (!pdata->io_endpoints) {
-		ERROR("Unable to allocate endpoints\n");
+		IIO_ERROR("Unable to allocate endpoints\n");
 		ret = -ENOMEM;
 		goto err_free_config_descriptor;
 	}
@@ -882,12 +945,12 @@ struct iio_context * usb_create_context(unsigned int bus,
 		ep->addr_out = iface->endpoint[i * 2 + 1].bEndpointAddress;
 		ep->pipe_id = i;
 
-		DEBUG("Couple %i with endpoints 0x%x / 0x%x\n", i,
+		IIO_DEBUG("Couple %i with endpoints 0x%x / 0x%x\n", i,
 				ep->addr_in, ep->addr_out);
 
 		ep->lock = iio_mutex_create();
 		if (!ep->lock) {
-			ERROR("Unable to create mutex\n");
+			IIO_ERROR("Unable to create mutex\n");
 			ret = -ENOMEM;
 			goto err_free_endpoints;
 		}
@@ -896,7 +959,7 @@ struct iio_context * usb_create_context(unsigned int bus,
 	pdata->ctx = usb_ctx;
 	pdata->hdl = hdl;
 	pdata->timeout_ms = DEFAULT_TIMEOUT_MS;
-	pdata->interface = interface;
+	pdata->intrfc = intrfc;
 
 	ret = usb_io_context_init(&pdata->io_ctx);
 	if (ret)
@@ -909,14 +972,14 @@ struct iio_context * usb_create_context(unsigned int bus,
 	ret = usb_reset_pipes(pdata);
 	if (ret) {
 		iio_strerror(-ret, err_str, sizeof(err_str));
-		ERROR("Failed to reset pipes: %s\n", err_str);
+		IIO_ERROR("Failed to reset pipes: %s\n", err_str);
 		goto err_io_context_exit;
 	}
 
 	ret = usb_open_pipe(pdata, 0);
 	if (ret) {
 		iio_strerror(-ret, err_str, sizeof(err_str));
-		ERROR("Failed to open control pipe: %s\n", err_str);
+		IIO_ERROR("Failed to open control pipe: %s\n", err_str);
 		goto err_io_context_exit;
 	}
 
@@ -932,12 +995,12 @@ struct iio_context * usb_create_context(unsigned int bus,
 	ctx->ops = &usb_ops;
 	ctx->pdata = pdata;
 
-	for (i = 0; i < ctx->nb_devices; i++) {
-		struct iio_device *dev = ctx->devices[i];
+	for (i = 0; i < iio_context_get_devices_count(ctx); i++) {
+		struct iio_device *dev = iio_context_get_device(ctx, i);
 
 		dev->pdata = zalloc(sizeof(*dev->pdata));
 		if (!dev->pdata) {
-			ERROR("Unable to allocate memory\n");
+			IIO_ERROR("Unable to allocate memory\n");
 			ret = -ENOMEM;
 			goto err_context_destroy;
 		}
@@ -966,8 +1029,7 @@ err_free_endpoints:
 	for (i = 0; i < pdata->nb_ep_couples; i++)
 		if (pdata->io_endpoints[i].lock)
 			iio_mutex_destroy(pdata->io_endpoints[i].lock);
-	if (pdata->io_endpoints)
-		free(pdata->io_endpoints);
+	free(pdata->io_endpoints);
 err_free_config_descriptor:
 	libusb_free_config_descriptor(conf_desc);
 err_libusb_close:
@@ -978,8 +1040,6 @@ err_destroy_iiod_client:
 	iiod_client_destroy(pdata->iiod_client);
 err_destroy_ep_mutex:
 	iio_mutex_destroy(pdata->ep_lock);
-err_destroy_mutex:
-	iio_mutex_destroy(pdata->lock);
 err_free_pdata:
 	free(pdata);
 err_set_errno:
@@ -989,58 +1049,97 @@ err_set_errno:
 
 struct iio_context * usb_create_context_from_uri(const char *uri)
 {
-	long bus, address, interface;
+	long bus, address, intrfc;
 	char *end;
 	const char *ptr;
+	/* keep MSVS happy by setting these to NULL */
+	struct iio_scan_context *scan_ctx = NULL;
+	struct iio_context_info **info = NULL;
+	bool scan = false;
 
 	if (strncmp(uri, "usb:", sizeof("usb:") - 1) != 0)
 		goto err_bad_uri;
 
 	ptr = (const char *) ((uintptr_t) uri + sizeof("usb:") - 1);
+
+	/* if uri is just "usb:" that means search for the first one */
+	if (!*ptr) {
+		ssize_t ret;
+
+		scan_ctx = iio_create_scan_context("usb", 0);
+		if (!scan_ctx) {
+			errno = ENOMEM;
+			goto err_bad_uri;
+		}
+
+		ret = iio_scan_context_get_info_list(scan_ctx, &info);
+		if (ret < 0) {
+			iio_scan_context_destroy(scan_ctx);
+			errno = ENOMEM;
+			goto err_bad_uri;
+		}
+		scan = true;
+		if (ret == 0 || ret > 1) {
+			errno = ENXIO;
+			goto err_bad_uri;
+		}
+		ptr = iio_context_info_get_uri(info[0]);
+		ptr += sizeof("usb:") - 1;
+	}
+
 	if (!isdigit(*ptr))
 		goto err_bad_uri;
 
+	errno = 0;
 	bus = strtol(ptr, &end, 10);
-	if (ptr == end || *end != '.')
+	if (ptr == end || *end != '.' || errno == ERANGE || bus < 0 || bus > UINT8_MAX)
 		goto err_bad_uri;
 
 	ptr = (const char *) ((uintptr_t) end + 1);
 	if (!isdigit(*ptr))
 		goto err_bad_uri;
 
+	errno = 0;
 	address = strtol(ptr, &end, 10);
-	if (ptr == end)
+	if (ptr == end || errno == ERANGE || address < 0 || address > UINT8_MAX)
 		goto err_bad_uri;
 
 	if (*end == '\0') {
-		interface = 0;
+		intrfc = 0;
 	} else if (*end == '.') {
 		ptr = (const char *) ((uintptr_t) end + 1);
 		if (!isdigit(*ptr))
 			goto err_bad_uri;
 
-		interface = strtol(ptr, &end, 10);
-		if (ptr == end || *end != '\0')
+		errno = 0;
+		intrfc = strtol(ptr, &end, 10);
+		if (ptr == end || *end != '\0' || errno == ERANGE || intrfc < 0 || intrfc > UINT8_MAX)
 			goto err_bad_uri;
 	} else {
 		goto err_bad_uri;
 	}
 
-	if (bus < 0 || address < 0 || interface < 0)
-		goto err_bad_uri;
-
+	if (scan) {
+		iio_context_info_list_free(info);
+		iio_scan_context_destroy(scan_ctx);
+	}
 	return usb_create_context((unsigned int) bus,
-			(unsigned int) address, (unsigned int) interface);
+			(uint16_t) address, (uint16_t) intrfc);
 
 err_bad_uri:
-	ERROR("Bad URI: \'%s\'\n", uri);
-	errno = EINVAL;
+	if (scan) {
+		iio_context_info_list_free(info);
+		iio_scan_context_destroy(scan_ctx);
+	} else
+		errno = EINVAL;
+
+	IIO_ERROR("Bad URI: \'%s\'\n", uri);
 	return NULL;
 }
 
 static int usb_fill_context_info(struct iio_context_info *info,
 		struct libusb_device *dev, struct libusb_device_handle *hdl,
-		unsigned int interface)
+		unsigned int intrfc)
 {
 	struct libusb_device_descriptor desc;
 	char manufacturer[64], product[64], serial[64];
@@ -1053,7 +1152,7 @@ static int usb_fill_context_info(struct iio_context_info *info,
 
 	iio_snprintf(uri, sizeof(uri), "usb:%d.%d.%u",
 		libusb_get_bus_number(dev), libusb_get_device_address(dev),
-		interface);
+		intrfc);
 
 	if (desc.iManufacturer == 0) {
 		manufacturer[0] = '\0';
@@ -1101,65 +1200,99 @@ static int usb_fill_context_info(struct iio_context_info *info,
 	return 0;
 }
 
-struct iio_scan_backend_context {
-	libusb_context *ctx;
-};
-
-struct iio_scan_backend_context * usb_context_scan_init(void)
+static int parse_vid_pid(const char *vid_pid, uint16_t *vid, uint16_t *pid)
 {
-	struct iio_scan_backend_context *ctx;
-	int ret;
+	unsigned long val;
+	char *ptr;
 
-	ctx = malloc(sizeof(*ctx));
-	if (!ctx) {
-		errno = ENOMEM;
-		return NULL;
-	}
+	/*
+	 * vid_pid string must be either:
+	 * - NULL: scan everything,
+	 * - "vid:*": scan all devices with the given VID,
+	 * - "vid:pid": scan the device with the given VID/PID.
+	 * IDs are given in hexadecimal, and the 0x prefix is not required.
+	 */
 
-	ret = libusb_init(&ctx->ctx);
-	if (ret) {
-		free(ctx);
-		errno = (int) libusb_to_errno(ret);
-		return NULL;
-	}
+	*vid = 0;
+	*pid = 0;
 
-	return ctx;
+	if (!vid_pid)
+		return 0;
+
+	val = strtoul(vid_pid, &ptr, 16);
+	if (ptr == vid_pid || val > 0xFFFF || *ptr != ':')
+		return -EINVAL;
+
+	*vid = (uint16_t) val;
+
+	vid_pid = ptr + 1;
+
+	if (*vid_pid == '*')
+		return vid_pid[1] == '\0' ? 0 : -EINVAL;
+
+	val = strtoul(vid_pid, &ptr, 16);
+	if (ptr == vid_pid || val > 0xFFFF || *ptr != '\0')
+		return -EINVAL;
+
+	*pid = (uint16_t) val;
+
+	return 0;
 }
 
-void usb_context_scan_free(struct iio_scan_backend_context *ctx)
+int usb_context_scan(struct iio_scan_result *scan_result, const char *args)
 {
-	libusb_exit(ctx->ctx);
-	free(ctx);
-}
-
-int usb_context_scan(struct iio_scan_backend_context *ctx,
-		struct iio_scan_result *scan_result)
-{
-	struct iio_context_info **info;
+	struct iio_context_info *info;
 	libusb_device **device_list;
+	libusb_context *ctx;
+	uint16_t vid, pid;
 	unsigned int i;
 	int ret;
 
-	ret = libusb_get_device_list(ctx->ctx, &device_list);
+	ret = parse_vid_pid(args, &vid, &pid);
+	if (ret)
+		return ret;
+
+	ret = libusb_init(&ctx);
 	if (ret < 0)
 		return -(int) libusb_to_errno(ret);
 
+	ret = (int) libusb_get_device_list(ctx, &device_list);
+	if (ret < 0) {
+		ret = -(int) libusb_to_errno(ret);
+		goto cleanup_libusb_exit;
+	}
+
 	for (i = 0; device_list[i]; i++) {
 		struct libusb_device_handle *hdl;
 		struct libusb_device *dev = device_list[i];
-		unsigned int interface = 0;
+		unsigned int intrfc = 0;
+		struct libusb_device_descriptor device_descriptor;
+
+		/* If we are given a pid or vid, use that to qualify for things,
+		 * this avoids open/closing random devices & potentially locking
+		 * (blocking them) from other applications
+		 */
+		if(vid || pid) {
+			ret = libusb_get_device_descriptor(dev, &device_descriptor);
+			if (ret)
+				continue;
+			if (vid && vid != device_descriptor.idVendor)
+				continue;
+			if (pid && pid != device_descriptor.idProduct)
+				continue;
+		}
 
 		ret = libusb_open(dev, &hdl);
 		if (ret)
 			continue;
 
-		if (!iio_usb_match_device(dev, hdl, &interface)) {
-			info = iio_scan_result_add(scan_result, 1);
+		if (!iio_usb_match_device(dev, hdl, &intrfc)) {
+			info = iio_scan_result_add(scan_result);
 			if (!info)
 				ret = -ENOMEM;
 			else
-				ret = usb_fill_context_info(*info, dev, hdl,
-						interface);
+				ret = usb_fill_context_info(info, dev, hdl,
+							    intrfc);
 		}
 
 		libusb_close(hdl);
@@ -1171,5 +1304,7 @@ int usb_context_scan(struct iio_scan_backend_context *ctx,
 
 cleanup_free_device_list:
 	libusb_free_device_list(device_list, true);
+cleanup_libusb_exit:
+	libusb_exit(ctx);
 	return ret;
 }
